#!/usr/bin/env python3

"""
Additional commands to add to the CLI beyond the OpenAPI spec.
"""
from __future__ import print_function
import functools
import operator
import os
import sys
import time

import click
import requests
import webbrowser

import civis
from civis.io import file_to_civis, civis_to_file


# From http://patorjk.com/software/taag/#p=display&f=3D%20Diagonal&t=CIVIS
_CIVIS_ASCII_ART = r"""
  ,----..     ,---,               ,---,  .--.--.
 /   /   \ ,`--.' |       ,---.,`--.' | /  /    '.
|   :     :|   :  :      /__./||   :  :|  :  /`. /
.   |  ;. /:   |  ' ,---.;  ; |:   |  ';  |  |--`
.   ; /--` |   :  |/___/ \  | ||   :  ||  :  ;_
;   | ;    '   '  ;\   ;  \ ' |'   '  ; \  \    `.
|   : |    |   |  | \   \  \: ||   |  |  `----.   \
.   | '___ '   :  ;  ;   \  ' .'   :  ;  __ \  \  |
'   ; : .'||   |  '   \   \   '|   |  ' /  /`--'  /
'   | '/  :'   :  |    \   `  ;'   :  |'--'.     /
|   :    / ;   |.'      :   \ |;   |.'   `--'---'
 \   \ .'  '---'         '---" '---'
  `---`
"""
_FOLLOW_LOG_NOTE = '''

Outputs job run logs in the format: "datetime message\\n" where
datetime is in ISO8601 format, like "2020-02-14T20:28:18.722Z".
If the job is still running, this command will continue outputting logs
until the run is done and then exit. If the run is already finished, it
will output all the logs from that run and then exit.

NOTE: This command could miss some log entries from a currently-running
job. It does not re-fetch logs that might have been saved out of order, to
preserve the chronological order of the logs and without duplication.
'''
_FOLLOW_POLL_INTERVAL_SEC = 3


@click.command('upload')
@click.argument('path')
@click.option('--name', type=str, default=None,
              help="A name for the Civis File (defaults to the base file name")
@click.option('--expires-at', type=str, default=None,
              help="The date and time the file will expire "
                   "(ISO-8601 format, e.g., \"2017-01-15\" or "
                   "\"2017-01-15T15:25:10Z\"). "
                   "Set \"never\" for the file to not expire."
                   "The default is the default in Civis (30 days).")
def files_upload_cmd(path, name, expires_at):
    """Upload a local file to Civis and get back the File ID."""

    if name is None:
        name = os.path.basename(path)

    if expires_at is None:
        # Use the default in Civis platform (30 days).
        expires_kwarg = {}
    elif expires_at.lower() == "never":
        expires_kwarg = {"expires_at": None}
    else:
        expires_kwarg = {"expires_at": expires_at}

    with open(path, 'rb') as f:
        file_id = file_to_civis(f, name=name, **expires_kwarg)
    print(file_id)


@click.command('download')
@click.argument('file_id', type=int)
@click.argument('path')
def files_download_cmd(file_id, path):
    """Download a Civis File to a specified local path."""
    with open(path, 'wb') as f:
        civis_to_file(file_id, f)


@click.command('sql')
@click.option('--dbname', '-d', type=str, required=True,
              help='Execute the query on this Civis Platform database')
@click.option('--command', '-c', type=str, default=None,
              help='Execute a single input command string')
@click.option('--filename', '-f', type=click.Path(exists=True),
              help='Execute a query read from the given file')
@click.option('--output', '-o', type=click.Path(),
              help='Download query results to this file')
@click.option('--quiet', '-q', is_flag=True, help='Suppress screen output')
@click.option('-n', type=int, default=100,
              help="Display up to this many rows of the result. Max 100.")
def sql_cmd(dbname, command, filename, output, quiet, n):
    """\b Execute a SQL query in Civis Platform

    If neither a command nor an input file is specified, read
    the SQL command from stdin.
    If writing to an output file, use a Civis SQL script and write the
    entire query output to the specified file.
    If not writing to an output file, use a Civis Query, and return a
    preview of the results, up to a maximum of 100 rows.
    """
    if filename:
        with open(filename, 'rt') as f:
            sql = f.read()
    elif not command:
        # Read the SQL query from user input. This also allows use of a heredoc
        lines = []
        while True:
            try:
                _i = input()
            except (KeyboardInterrupt, EOFError):
                # The end of a heredoc produces an EOFError.
                break
            if not _i:
                break
            else:
                lines.append(_i)
        sql = '\n'.join(lines)
    else:
        sql = command

    if not sql:
        # If the user didn't enter a query, exit.
        if not quiet:
            print('ERROR: Did not receive a SQL query.', file=sys.stderr)
        return

    if not quiet:
        print('\nExecuting query...', file=sys.stderr)
    if output:
        fut = civis.io.civis_to_csv(output, sql, database=dbname)
        fut.result()  # Block for completion and raise exceptions if any
        if not quiet:
            print("Downloaded the result of the query to %s." % output,
                  file=sys.stderr)
    else:
        fut = civis.io.query_civis(sql, database=dbname,
                                   preview_rows=n, polling_interval=3)
        cols = fut.result()['result_columns']
        rows = fut.result()['result_rows']
        if not quiet:
            print('...Query complete.\n', file=sys.stderr)
            print(_str_table_result(cols, rows))


def _str_table_result(cols, rows):
    """Turn a Civis Query result into a readable table."""
    str_rows = [['' if _v is None else _v for _v in row] for row in rows]
    # Determine the maximum width of each column.
    # First find the width of each element in each row, then find the max
    # width in each position.
    max_len = functools.reduce(
        lambda x, y: [max(z) for z in zip(x, y)],
        [[len(_v) for _v in _r] for _r in [cols] + str_rows])

    header_str = " | ".join("{0:<{width}}".format(_v, width=_l)
                            for _l, _v in zip(max_len, cols))
    tb_strs = [header_str, len(header_str) * '-']
    for row in str_rows:
        tb_strs.append(" | ".join("{0:>{width}}".format(_v, width=_l)
                                  for _l, _v in zip(max_len, row)))
    return '\n'.join(tb_strs)


@click.command(
    'follow-log',
    help='Output live log from the most recent job run.' + _FOLLOW_LOG_NOTE)
@click.argument('id', type=int)
def jobs_follow_log(id):
    client = civis.APIClient()
    runs = client.jobs.list_runs(id, limit=1, order='id', order_dir='desc')
    if not runs:
        raise click.ClickException('No runs found for that job ID.')
    run_id = runs[0].id
    print('Run ID: ' + str(run_id))
    _jobs_follow_run_log(id, run_id)


@click.command(
    'follow-run-log',
    help='Output live run log.' + _FOLLOW_LOG_NOTE)
@click.argument('id', type=int)
@click.argument('run_id', type=int)
def jobs_follow_run_log(id, run_id):
    _jobs_follow_run_log(id, run_id)


def _jobs_follow_run_log(id, run_id):
    client = civis.APIClient(return_type='raw')
    local_max_log_id = 0
    continue_polling = True

    while continue_polling:
        # This call gets all available log messages since last_id up to
        # the page size, ordered by log ID. We leave it to Platform to decide
        # the best page size.
        response = client.jobs.list_runs_logs(id, run_id,
                                              last_id=local_max_log_id)
        if 'civis-max-id' in response.headers:
            remote_max_log_id = int(response.headers['civis-max-id'])
        else:
            # Platform hasn't seen any logs at all yet
            remote_max_log_id = None
        logs = response.json()
        if logs:
            local_max_log_id = max(log['id'] for log in logs)
            logs.sort(key=operator.itemgetter('createdAt', 'id'))
        for log in logs:
            print(' '.join((log['createdAt'], log['message'].rstrip())))
        # if output is a pipe, write the buffered output immediately:
        sys.stdout.flush()

        log_finished = response.headers['civis-cache-control'] != 'no-store'
        if remote_max_log_id is None:
            remote_has_more_logs_to_get_now = False
        elif local_max_log_id == remote_max_log_id:
            remote_has_more_logs_to_get_now = False
            if log_finished:
                continue_polling = False
        else:
            remote_has_more_logs_to_get_now = True

        if continue_polling and not remote_has_more_logs_to_get_now:
            time.sleep(_FOLLOW_POLL_INTERVAL_SEC)


@click.command('download')
@click.argument('notebook_id', type=int)
@click.argument('path')
def notebooks_download_cmd(notebook_id, path):
    """Download a notebook to a specified local path."""
    client = civis.APIClient()
    info = client.notebooks.get(notebook_id)
    response = requests.get(info['notebook_url'], stream=True)
    response.raise_for_status()
    chunk_size = 32 * 1024
    chunked = response.iter_content(chunk_size)
    with open(path, 'wb') as f:
        for lines in chunked:
            f.write(lines)


@click.command('new')
@click.argument('language', type=click.Choice(['python3', 'python2', 'r']),
                default='python3')
@click.option('--mem', type=int, default=None,
              help='Memory allocated for this notebook in MiB.')
@click.option('--cpu', type=int, default=None,
              help='CPU available for this notebook in 1/1000 of a core.')
def notebooks_new_cmd(language='python3', mem=None, cpu=None):
    """Create a new notebook and open it in the browser."""
    client = civis.APIClient()
    kwargs = {'memory': mem, 'cpu': cpu}
    kwargs = {k: v for k, v in kwargs.items() if v is not None}
    new_nb = client.notebooks.post(language=language, **kwargs)
    print("Created new {language} notebook with ID {id} in Civis Platform"
          " (https://platform.civisanalytics.com/#/notebooks/{id})."
          .format(language=language, id=new_nb.id))
    _notebooks_up(new_nb.id)
    _notebooks_open(new_nb.id)


@click.command('up')
@click.argument('notebook_id', type=int)
@click.option('--mem', type=int, default=None,
              help='Memory allocated for this notebook in MiB.')
@click.option('--cpu', type=int, default=None,
              help='CPU available for this notebook in 1/1000 of a core.')
def notebooks_up(notebook_id, mem=None, cpu=None):
    """Start an existing notebook and open it in the browser."""
    client = civis.APIClient()
    kwargs = {'memory': mem, 'cpu': cpu}
    kwargs = {k: v for k, v in kwargs.items() if v is not None}
    client.notebooks.patch(notebook_id, **kwargs)
    _notebooks_up(notebook_id)
    _notebooks_open(notebook_id)


def _notebooks_up(notebook_id):
    client = civis.APIClient()
    return client.notebooks.post_deployments(notebook_id)


@click.command('down')
@click.argument('notebook_id', type=int)
def notebooks_down(notebook_id):
    """Shut down a running notebook."""
    client = civis.APIClient()
    nb = client.notebooks.get(notebook_id)
    state = nb['most_recent_deployment']['state']
    if state not in ['running', 'pending']:
        print('Notebook is in state "{}" and can\'t be stopped.'.format(state))
    deployment_id = nb['most_recent_deployment']['deploymentId']
    client.notebooks.delete_deployments(notebook_id, deployment_id)


@click.command('open')
@click.argument('notebook_id', type=int)
def notebooks_open(notebook_id):
    """Open an existing notebook in the browser."""
    _notebooks_open(notebook_id)


def _notebooks_open(notebook_id):
    url = 'https://platform.civisanalytics.com/#/notebooks/{}?fullscreen=true'
    url = url.format(notebook_id)
    webbrowser.open(url, new=2, autoraise=True)


@click.command('civis', help="Print Civis")
def civis_ascii_art():
    print(_CIVIS_ASCII_ART)



import pytest
import civis

from civis.response import Response
from civis.tests.mocks import create_client_mock
from civis.utils._jobs import logging


def create_mock_client_with_job():
    mock_client = create_client_mock()

    job_response = Response({"id": 1, "name": "test"})
    run_post_response = Response({"id": 1})
    run_get_response = Response(
        {
            "id": run_post_response.id,
            "state": "succeeded",
            "is_cancel_requested": False,
            "error": None,
            "custom_id": run_post_response.id,
        }
    )
    mock_client.scripts.post_custom.return_value = job_response
    mock_client.scripts.post_custom_runs.return_value = run_post_response
    mock_client.scripts.get_custom_runs.return_value = run_get_response
    return mock_client


@pytest.fixture
def mock_client_single_json_output():
    mock_client = create_mock_client_with_job()
    mock_output = [
        Response(
            {
                "object_type": "JSONValue",
                "object_id": 10,
                "name": "output",
                "value": {"a": 1},
            }
        )
    ]
    mock_client.scripts.list_custom_runs_outputs.return_value = mock_output
    return mock_client


@pytest.fixture
def mock_client_multiple_json_output():
    mock_client = create_mock_client_with_job()
    mock_output = [
        Response(
            {
                "object_type": "JSONValue",
                "object_id": 10,
                "name": "output",
                "value": {"a": 1},
            }
        ),
        Response(
            {
                "object_type": "JSONValue",
                "object_id": 11,
                "name": "output2",
                "value": {"b": 2},
            }
        ),
    ]
    mock_client.scripts.list_custom_runs_outputs.return_value = mock_output
    return mock_client


@pytest.fixture
def mock_client_no_json_output():
    mock_client = create_mock_client_with_job()
    mock_output = [
        Response(
            {
                "object_type": int,
                "object_id": 10,
                "name": "output",
                "value": 100,
            }
        )
    ]
    mock_client.scripts.list_custom_runs_outputs.return_value = mock_output
    return mock_client


def test_run_template_json_output_fileids_returned(
    mock_client_single_json_output
):
    template_id = 1
    args = {"arg": 1}
    result = civis.utils.run_template(
        template_id, arguments=args, client=mock_client_single_json_output
    )
    assert result == {"output": 10}


def test_run_template_no_json_output_fileids_returned(
    mock_client_no_json_output
):
    template_id = 1
    args = {"arg": 1}
    result = civis.utils.run_template(
        template_id, arguments=args, client=mock_client_no_json_output
    )
    assert result == {"output": 10}


def test_run_template_multiple_json_output_fileids_returned(
    mock_client_multiple_json_output
):
    template_id = 1
    args = {"arg": 1}
    result = civis.utils.run_template(
        template_id, arguments=args, client=mock_client_multiple_json_output
    )
    assert result == {"output": 10, "output2": 11}


def test_run_template_json_returned(mock_client_single_json_output):
    template_id = 1
    args = {"arg": 1}
    result = civis.utils.run_template(
        template_id,
        arguments=args,
        JSONValue=True,
        client=mock_client_single_json_output,
    )
    assert result == {"a": 1}


def test_run_template_many_json_outputs(
    caplog, mock_client_multiple_json_output
):
    template_id = 1
    args = {"arg": 1}
    result = civis.utils.run_template(
        template_id,
        arguments=args,
        JSONValue=True,
        client=mock_client_multiple_json_output,
    )
    log_result = caplog.record_tuples
    assert log_result == [
        (
            "civis.utils._jobs",
            logging.WARNING,
            "More than 1 JSON output for template {} "
            "-- returning only the first one.".format(template_id),
        )
    ]
    assert result == {"a": 1}


def test_run_template_when_no_json_output(caplog, mock_client_no_json_output):
    template_id = 1
    args = {"arg": 1}
    result = civis.utils.run_template(
        template_id,
        arguments=args,
        JSONValue=True,
        client=mock_client_no_json_output,
    )
    log_result = caplog.record_tuples
    assert log_result == [
        (
            "civis.utils._jobs",
            logging.WARNING,
            "No JSON output for template {}".format(template_id),
        )
    ]
    assert result is None



#!/usr/bin/env python3

"""
Civis Command Line Interface

This is based on https://github.com/zalando/openapi-cli-client,
which has an Apache 2.0 License:
https://github.com/zalando/openapi-cli-client/blob/master/LICENSE
"""
from __future__ import print_function


import calendar
from collections import OrderedDict
from functools import partial
import json
import logging
import os
import re
import sys
import time
from warnings import warn

import click
from jsonref import JsonRef
import yaml
from requests import Request
from civis.cli._cli_commands import (
    civis_ascii_art, files_download_cmd, files_upload_cmd,
    jobs_follow_log, jobs_follow_run_log, notebooks_download_cmd,
    notebooks_new_cmd, notebooks_up, notebooks_down, notebooks_open, sql_cmd)
from civis.resources import get_api_spec, CACHED_SPEC_PATH
from civis.resources._resources import parse_method_name
from civis._utils import open_session, retry_request, MAX_RETRIES


_REPLACEABLE_COMMAND_CHARS = re.compile(r'[^A-Za-z0-9]+')
_BASE_API_URL = "https://api.civisanalytics.com"
CLI_USER_AGENT = 'civis-cli'


class YAMLParamType(click.ParamType):
    """
    A click parameter type for YAML/JSON files.
    See http://click.pocoo.org/5/parameters/#implementing-custom-types.
    """

    name = 'yamlpath'

    def convert(self, value, param, ctx):
        if value is None:
            return value

        try:
            with open(value) as f:
                result = yaml.safe_load(f)
                return result
        except Exception:
            self.fail("Could not load YAML from path: %s" % value, param, ctx)


YAML = YAMLParamType()


# A map from OpenAPI type strings to python types.
# See https://github.com/OAI/OpenAPI-Specification/blob/master/versions/2.0.md#data-types.  # noqa: E501
_TYPE_MAP = {
    "integer": int,
    "number": float,
    "boolean": bool,
    "string": str,
    "object": YAML,
    "array": YAML
}


def get_api_key():
    try:
        return os.environ["CIVIS_API_KEY"]
    except KeyError:
        print("You must set the CIVIS_API_KEY environment variable.",
              file=sys.stderr)
        sys.exit(1)


def get_base_api_url():
    return os.getenv('CIVIS_API_ENDPOINT') or _BASE_API_URL


def camel_to_snake(s):
    return re.sub(r'([A-Z]+)', r'_\1', s).lower()


def munge_name(s):
    """Replace non-alphanumeric characters with dashes."""
    result = _REPLACEABLE_COMMAND_CHARS.sub('-', s.lower()).strip('-')
    return result


def make_operation_name(path, method, resource_name):
    """Create an appropriate CLI command for an operation.

    Examples
    --------
    >>> make_operation_name('/scripts/r/{id}/runs/{run_id}', 'get', 'scripts')
    get-r-runs
    """
    path = path.lower().lstrip('/')

    # Remove resource prefix. Note that the path name for some operations is
    # just the resource name (e.g., /databases).
    if path.startswith(resource_name):
        path = path[len(resource_name):].strip("-")

    name = parse_method_name(method, path).replace("_", "-")
    return name


def param_case_map(param_names):
    # Return a map from snake_case parameter names (for click) to camelCase
    # versions for the API. We need this mapping because the API
    # expects/prefers camelCase but click prefers snake_case.
    # We could have make a snake_to_camel function, but ensuring that that
    # was in sync with camel_to_snake might be difficult because there are some
    # edge cases like provideAPIKey <-> provide_api_key.
    result = {camel_to_snake(k): k for k in param_names}
    return result


def invoke(method, path, op, *args, **kwargs):
    """
    If json_output is in `kwargs` then the output is json. Otherwise, it is
    yaml.
    """
    # Remove None b/c click passes everything in as None if it's not set.
    kwargs = {k: v for k, v in kwargs.items()
              if v is not None}
    json_output = kwargs.pop('json_output', False)

    # Construct the body of the request.
    body = {}
    body_params = [p for p in op['parameters'] if p['in'] == 'body']
    if body_params:
        assert len(body_params) == 1  # There can be only one body parameter.
        props = body_params[0]['schema']['properties']
        param_map = param_case_map(props.keys())
        body = {param_map[k]: v for k, v in kwargs.items()
                if k in param_map}

    # Construct the query part of the request.
    query_names = {p['name'] for p in op['parameters'] if p['in'] == 'query'}
    param_map = param_case_map(query_names)
    query = {param_map[k]: v for k, v in kwargs.items()
             if k in param_map}

    # Make the request.
    request_info = dict(
        params=query,
        json=body,
        url=get_base_api_url() + path.format(**kwargs),
        method=method
    )
    with open_session(get_api_key(), user_agent=CLI_USER_AGENT) as sess:
        request = Request(**request_info)
        pre_request = sess.prepare_request(request)
        response = retry_request(method, pre_request, sess, MAX_RETRIES)

    # Print the response to stderr and set exit code to 1 if there was an error
    output_file = sys.stdout
    exit_code = 0
    if not (200 <= response.status_code < 300):
        output_file = sys.stderr
        exit_code = 1

    # Print the output, if there is any.
    # For commands such as DELETE /scripts/containers/{script_id}/runs/{id},
    # response ends up being " " here.
    try:
        if json_output:
            json.dump(response.json(), output_file)
        else:
            yaml.safe_dump(response.json(), output_file,
                           default_flow_style=False)
        output_file.flush()

    # json throws a ValueError if it is passed a blank string to load.
    except ValueError as e:
        # If the message was not blank, print an error message.
        # Otherwise, do nothing.
        if response.text.strip():
            print("Error parsing response: {}".format(e), file=sys.stderr)

    sys.exit(exit_code)


def retrieve_spec_dict(api_version="1.0"):
    """Retrieve the API specification from a cached version or from Civis."""

    refresh_spec = True
    now_timestamp = calendar.timegm(time.gmtime())

    try:
        # If the cached spec is from the last 24 hours, use it.
        modified_time = os.path.getmtime(CACHED_SPEC_PATH)
        if now_timestamp - modified_time < 24 * 3600:
            refresh_spec = False
            with open(CACHED_SPEC_PATH) as f:
                spec_dict = json.load(f, object_pairs_hook=OrderedDict)
    except (FileNotFoundError, ValueError):
        # If the file doesn't exist or we can't parse it, just keep going.
        refresh_spec = True

    # Download the spec and cache it in the user's home directory.
    if refresh_spec:
        spec_dict = get_api_spec(get_api_key(), api_version=api_version,
                                 user_agent=CLI_USER_AGENT)
        with open(CACHED_SPEC_PATH, "w") as f:
            json.dump(spec_dict, f)
    return spec_dict


def add_extra_commands(cli):
    """Add useful commands that are not in the OpenAPI spec."""
    files_cmd = cli.commands['files']
    files_cmd.add_command(files_download_cmd)
    files_cmd.add_command(files_upload_cmd)
    notebooks_cmd = cli.commands['notebooks']
    notebooks_cmd.add_command(notebooks_download_cmd)
    notebooks_cmd.add_command(notebooks_new_cmd)
    notebooks_cmd.add_command(notebooks_up)
    notebooks_cmd.add_command(notebooks_down)
    notebooks_cmd.add_command(notebooks_open)
    jobs_cmd = cli.commands['jobs']
    jobs_cmd.add_command(jobs_follow_log)
    jobs_cmd.add_command(jobs_follow_run_log)
    cli.add_command(civis_ascii_art)

    cli.add_command(sql_cmd)


def configure_log_level():
    if os.getenv('CIVIS_LOG_LEVEL'):
        logging.basicConfig(level=os.getenv('CIVIS_LOG_LEVEL'))


def generate_cli():
    configure_log_level()
    spec = retrieve_spec_dict()

    # Replace references in the spec so that we don't have to worry about them
    # when making the CLI.
    spec = JsonRef.replace_refs(spec)

    cli = click.Group()

    # Iterate through top-level resources (e.g., Scripts, Files, Models).
    groups = {}
    for path, path_dict in spec['paths'].items():
        resource = path.strip('/').split('/')[0]
        grp = groups.get(resource)
        if grp is None:
            grp = click.Group(munge_name(resource),
                              short_help='Manage {}'.format(resource))
            cli.add_command(grp)
            groups[resource] = grp

        add_path_commands(path, path_dict, grp, resource)

    add_extra_commands(cli)

    return cli


def add_path_commands(path, path_dict, grp, resource):
    """Add commands for the given resource."""

    for method, op_dict in path_dict.items():

        # Set up a command with the appropriate name and help info.
        name = make_operation_name(path, method, resource)
        summary = op_dict.get('summary', '')
        op_help = summary
        description = op_dict.get('description', '')
        if description:
            op_help += '\n\n' + description
        op_help += '\n\n' + path
        callback = partial(invoke, method=method, path=path, op=op_dict)
        cmd = click.Command(name,
                            callback=callback,
                            short_help=summary,
                            help=op_help)

        add_command_params(cmd, op_dict)

        # Add an option for JSON output.
        cmd.params.append(click.Option(['--json-output'], is_flag=True,
                                       help="output in JSON instead of YAML"))

        if cmd.name in grp.commands:
            warn('conflicting command name "%s" for path "%s"' %
                 (cmd.name, path))

        grp.add_command(cmd)


def add_command_params(cmd, op_dict):
    """Add parameters to the click command for an API operation."""

    # Extract properties of objects in body to make click params for them.
    parameters_orig = op_dict.get('parameters', [])
    parameters = []
    for p in parameters_orig:
        if p['in'] == 'body':
            body_params = []
            req = p['schema'].get('required', [])
            for prop_name, prop_info in p['schema']['properties'].items():
                p_new = dict(
                    name=camel_to_snake(prop_name),
                    type=prop_info['type'],
                    description=prop_info.get('description', ''),
                    required=req == 'all' or prop_name in req,
                )
                body_params.append(p_new)

            # Sort the parameters since they don't have an order as properties.
            body_params = sorted(body_params, key=lambda x: x['name'])

            parameters.extend(body_params)
        else:
            parameters.append(p)

    # Specify the parameters for this command.
    for p in parameters:
        param_type_spec = p.get('type', 'string')
        param_type = _TYPE_MAP[param_type_spec]
        description = p.get('description', '')

        if p['required']:
            cmd.help += "\n\n{} ({}) - {}".format(
                p['name'].upper(), param_type_spec, description)
            arg = click.Argument([p['name'].lower()],
                                 type=param_type)
            cmd.params.append(arg)
        else:
            arg = click.Option(['--' + munge_name(p['name'].lower())],
                               help=description,
                               type=param_type)
            cmd.params.append(arg)


def main():
    # Note: this needs to be its own function so that setup.py can make it an
    # entry point.
    cli = generate_cli()
    cli()


if __name__ == '__main__':
    main()



from collections import OrderedDict
import json
import os
from unittest import mock

import pytest

from civis.cli.__main__ import generate_cli, invoke, make_operation_name
from civis.cli._cli_commands import _str_table_result
from civis.tests import TEST_SPEC

THIS_DIR = os.path.dirname(os.path.realpath(__file__))


@mock.patch("civis.cli.__main__.add_extra_commands")
@mock.patch("civis.cli.__main__.retrieve_spec_dict")
def test_generate_cli_petstore(mock_retrieve_spec_dict,
                               mock_add_extra_commands):
    """Test loading the OpenAPI petstore example."""

    # From https://raw.githubusercontent.com/OAI/OpenAPI-Specification/4b1c1167b99844fd3ca19dc0055bbdb0c5eff094/examples/v2.0/json/petstore.json  # noqa: E501
    with open(os.path.join(THIS_DIR, "petstore.json")) as f:
        petstore_spec = json.load(f)

    mock_retrieve_spec_dict.return_value = petstore_spec
    cli = generate_cli()
    assert set(cli.commands.keys()) == {'pets'}
    assert set(cli.commands['pets'].commands.keys()) == {'list', 'post', 'get'}
    assert ({x.name for x in cli.commands['pets'].commands['list'].params} ==
            {'limit', 'json_output'})


@mock.patch("civis.cli.__main__.retrieve_spec_dict")
def test_generate_cli_civis(mock_retrieve_spec_dict):
    """Test loading the Civis API spec as of 2017-02-02."""
    with open(TEST_SPEC) as f:
        civis_spec = json.load(f, object_pairs_hook=OrderedDict)
    mock_retrieve_spec_dict.return_value = civis_spec

    with pytest.warns(None) as warn_rec:
        cli = generate_cli()
    assert len(warn_rec) == 0

    # Check a regular command.
    list_runs_cmd = cli.commands['scripts'].commands['list-containers-runs']
    expected_names = {'id', 'limit', 'page_num', 'order', 'order_dir',
                      'json_output'}
    assert {_.name for _ in list_runs_cmd.params} == expected_names
    assert list_runs_cmd.params[0].name == 'id'
    assert list_runs_cmd.params[0].required
    assert list_runs_cmd.params[1].name == 'limit'
    assert not list_runs_cmd.params[1].required

    # Check that the extra files upload command was added
    expected_names = {'path', 'name', 'expires_at'}
    files_upload_params = cli.commands['files'].commands['upload'].params
    assert {_.name for _ in files_upload_params} == expected_names

    # Check that the POST queries command, which uses an object in the body,
    # was parsed properly.
    pq_params = cli.commands['queries'].commands['post'].params
    expected_names = {
        'column_delimiter', 'compression', 'credential', 'database',
        'filename_prefix', 'include_header', 'interactive', 'preview_rows',
        'sql', 'unquoted', 'json_output', 'hidden'
    }
    assert {_.name for _ in pq_params} == expected_names
    for p in pq_params:
        if p.name == 'filename_prefix':
            assert not p.required
        if p.name == 'database':
            assert p.required


@mock.patch("civis.cli.__main__.open_session", autospec=True)
def test_blank_output(mock_session):
    """
    Test that endpoints that return blank results don't cause exceptions.
    """
    # The response object's json method will raise a ValueError when the output
    # is blank.
    session_context = mock_session.return_value.__enter__.return_value
    session_context.send.return_value.json.side_effect = ValueError()
    session_context.send.return_value.status_code = 200

    op = {"parameters": []}
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        invoke("WIBBLE", "/wobble/wubble", op)

    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code == 0


@mock.patch("civis.cli.__main__.open_session", autospec=True)
def test_failure_exit_code(mock_session):
    """
    Test that we return a nonzero exit code when the API request fails.
    """
    # first test that we get a zero exit code when the API request succeeds
    session_context = mock_session.return_value.__enter__.return_value
    session_context.send.return_value.json.side_effect = ValueError()
    session_context.send.return_value.status_code = 200

    op = {"parameters": []}

    with pytest.raises(SystemExit) as pytest_wrapped_e:
        invoke("WIBBLE", "/wobble/wubble", op)
    assert pytest_wrapped_e.value.code == 0

    # now test that we get a nonzero exit code when the API request fails
    session_context.send.return_value.status_code = 404

    with pytest.raises(SystemExit) as pytest_wrapped_e:
        invoke("WIBBLE", "/wobble/wubble", op)
    assert pytest_wrapped_e.value.code != 0


@mock.patch("civis.cli.__main__.open_session", autospec=True)
@mock.patch("civis.cli.__main__.Request", autospec=True)
def test_parameter_case(mock_request, mock_session):
    """
    Test that parameter names are sent in camelCase rather than snake_case.
    """
    api_response = {'key': 'value'}
    session_context = mock_session.return_value.__enter__.return_value
    session_context.send.return_value.json.return_value = api_response
    session_context.send.return_value.status_code = 200

    # To avoid needing CIVIS_API_KEY set in the environment.
    op = {"parameters": [{'name': 'firstParameter', 'in': 'query'},
                         {'name': 'secondParameter', 'in': 'query'}]}
    with pytest.raises(SystemExit):
        invoke("WIBBLE", "/wobble/wubble", op,
               first_parameter='a', second_parameter='b')

    mock_session.call_args[1]['user_agent'] = 'civis-cli'

    mock_request.assert_called_with(
        url='https://api.civisanalytics.com/wobble/wubble',
        json={},
        params={'firstParameter': 'a', 'secondParameter': 'b'},
        method='WIBBLE')


@pytest.mark.parametrize(
    "path,method,resource_name,exp",
    [('/imports/files/{id}/runs/{run_id}', 'get', 'imports', 'get-files-runs'),
     ('/aliases/{object_type}/{alias}', 'get', 'aliases', 'get-object-type'),
     ('/workflows/', 'get', 'workflows', 'list'),
     ('/results/{id}/grants', 'delete', 'results', 'delete-grants'),
     ]
)
def test_make_operation_name(path, method, resource_name, exp):
    assert make_operation_name(path, method, resource_name) == exp


def test_str_table_result():
    cols = ['a', 'snake!']
    rows = [['2', '3'], ['1.1', None]]

    out = _str_table_result(cols, rows)
    assert out == "a   | snake!\n------------\n  2 |      3\n1.1 |       "



from unittest import mock

import pytest

from civis.response import Response
from civis.ml import (list_models, put_models_shares_groups,
                      put_models_shares_users, delete_models_shares_groups,
                      delete_models_shares_users)
from civis.ml import _helper as helper
from civis.ml.tests import test_model
from civis.tests.mocks import create_client_mock


def test_list_models_bad_job_type():
    with pytest.raises(ValueError):
        list_models(job_type="fake")


def test_list_models():
    resp = [Response({'id': 2834, 'name': 'RFC model'})]
    m_client = create_client_mock()
    m_client.aliases.list.return_value = test_model.TEST_TEMPLATE_ID_ALIAS_OBJECTS  # noqa
    m_client.scripts.list_custom.return_value = resp
    out = list_models(job_type='train', client=m_client)
    assert out == resp

    out = list_models(job_type='predict', client=m_client)
    assert out == resp

    out = list_models(job_type=None, client=m_client)
    assert out == resp


def _create_share_model_client_mock(run_ids):
    m_client = create_client_mock()
    m_client.scripts.put_containers_shares_users.return_value = 'usershare'
    m_client.scripts.put_containers_shares_groups.return_value = 'groupshare'
    m_client.scripts.delete_containers_shares_users.return_value = 'userdel'
    m_client.scripts.delete_containers_shares_groups.return_value = 'groupdel'
    m_client.scripts.list_containers_runs.return_value = [
        Response({'id': _id}) for _id in run_ids]
    m_client.scripts.list_containers_runs_outputs.return_value = [
        Response({'object_id': 117, 'object_type': 'File', 'name': 'fname'}),
        Response({'object_id': 31, 'object_type': 'Project'}),
        Response({'object_id': 37, 'object_type': 'JSONValue'}),
    ]
    return m_client


def test_share_model_users():
    m_client = _create_share_model_client_mock([11])

    resp = helper._share_model(3, [7, 8], 'write', 'users', client=m_client,
                               send_shared_email=True)
    assert resp == 'usershare'
    m_client.scripts.put_containers_shares_users.assert_called_once_with(
        3, [7, 8], 'write', send_shared_email=True)
    m_client.files.put_shares_users.assert_called_once_with(
        117, [7, 8], 'write', send_shared_email=False)
    m_client.projects.put_shares_users.assert_called_once_with(
        31, [7, 8], 'write', send_shared_email=False)
    m_client.json_values.put_shares_users.assert_called_once_with(
        37, [7, 8], 'write', send_shared_email=False)


def test_share_model_groups():
    m_client = _create_share_model_client_mock([11])

    resp = helper._share_model(3, [7, 8], 'write', 'groups', client=m_client,
                               send_shared_email=True)
    assert resp == 'groupshare'
    m_client.scripts.put_containers_shares_groups.assert_called_once_with(
        3, [7, 8], 'write', send_shared_email=True)
    m_client.files.put_shares_groups.assert_called_once_with(
        117, [7, 8], 'write', send_shared_email=False)
    m_client.projects.put_shares_groups.assert_called_once_with(
        31, [7, 8], 'write', send_shared_email=False)
    m_client.json_values.put_shares_groups.assert_called_once_with(
        37, [7, 8], 'write', send_shared_email=False)


def test_share_model_tworuns():
    # Check that we grant permission on run outputs for each run
    m_client = _create_share_model_client_mock([11, 13])

    helper._share_model(3, [7, 8], 'write', 'users', client=m_client)

    m_client.scripts.put_containers_shares_users.assert_called_once_with(
        3, [7, 8], 'write')

    assert m_client.files.put_shares_users.call_count == 2
    assert m_client.projects.put_shares_users.call_count == 2
    assert m_client.json_values.put_shares_users.call_count == 2


def test_share_model_project_permissions():
    # Grant "write" permission on the internal project when
    # overall "read" permission is requested.
    m_client = _create_share_model_client_mock([11])

    helper._share_model(3, [7, 8], 'read', 'groups', client=m_client)

    m_client.projects.put_shares_groups.assert_called_once_with(
        31, [7, 8], 'write', send_shared_email=False)


def test_unshare_model_users():
    m_cl = _create_share_model_client_mock([11])

    resp = helper._unshare_model(3, 7, 'users', client=m_cl)
    assert resp == 'userdel'
    m_cl.scripts.delete_containers_shares_users.assert_called_once_with(3, 7)
    m_cl.files.delete_shares_users.assert_called_once_with(117, 7)
    m_cl.projects.delete_shares_users.assert_called_once_with(31, 7)
    m_cl.json_values.delete_shares_users.assert_called_once_with(37, 7)


def test_unshare_model_groups():
    m_cl = _create_share_model_client_mock([11])

    resp = helper._unshare_model(3, 7, 'groups', client=m_cl)
    assert resp == 'groupdel'
    m_cl.scripts.delete_containers_shares_groups.assert_called_once_with(3, 7)
    m_cl.files.delete_shares_groups.assert_called_once_with(117, 7)
    m_cl.projects.delete_shares_groups.assert_called_once_with(31, 7)
    m_cl.json_values.delete_shares_groups.assert_called_once_with(37, 7)


def test_unshare_model_tworuns():
    # Check that we grant permission on run outputs for each run
    m_cl = _create_share_model_client_mock([11, 13])

    helper._unshare_model(3, 7, 'users', client=m_cl)

    m_cl.scripts.delete_containers_shares_users.assert_called_once_with(3, 7)

    assert m_cl.files.delete_shares_users.call_count == 2
    assert m_cl.projects.delete_shares_users.call_count == 2
    assert m_cl.json_values.delete_shares_users.call_count == 2


@mock.patch('civis.ml._helper._share_model', autospec=True)
def test_put_models_shares_groups(mock_share):
    mock_share.return_value = 'retval'
    out = put_models_shares_groups(1, [7, 8], 'read')

    assert out == 'retval'
    mock_share.assert_called_once_with(1, [7, 8], 'read', entity_type='groups',
                                       client=None)


@mock.patch('civis.ml._helper._share_model', autospec=True)
def test_put_models_shares_users(mock_share):
    mock_share.return_value = 'retval'
    out = put_models_shares_users(1, [7, 8], 'read')

    assert out == 'retval'
    mock_share.assert_called_once_with(1, [7, 8], 'read', entity_type='users',
                                       client=None)


@mock.patch('civis.ml._helper._unshare_model', autospec=True)
def test_delete_models_shares_groups(m_unshare):
    m_unshare.return_value = 'retval'
    out = delete_models_shares_groups(1, 7)

    assert out == 'retval'
    m_unshare.assert_called_once_with(1, 7, entity_type='groups', client=None)


@mock.patch('civis.ml._helper._unshare_model', autospec=True)
def test_delete_models_shares_users(m_unshare):
    m_unshare.return_value = 'retval'
    out = delete_models_shares_users(1, 7)

    assert out == 'retval'
    m_unshare.assert_called_once_with(1, 7, entity_type='users', client=None)



import os

from vcr import persist
from vcr.serialize import CASSETTE_FORMAT_VERSION
from vcr.serializers import compat
from vcr_unittest import VCRTestCase

# The "GENERATE_TEST" environment variable indicates that
# we're recording new cassettes.
if os.environ.get('GENERATE_TESTS'):
    # Use default polling intervals if generating new tests.
    POLL_INTERVAL = None
else:
    # Speed through calls in pre-recorded VCR cassettes.
    POLL_INTERVAL = 0.00001


def cassette_dir():
    testdir = os.path.dirname(os.path.abspath(__file__))
    return os.path.join(testdir, 'cassettes')


# monkey patches the persist.serialize function in vcrpy
# https://github.com/kevin1024/vcrpy/issues/273


def github_serialize(cassette_dict, serializer):
    interactions = ([{
        'request': compat.convert_to_unicode(request._to_dict()),
        'response': compat.convert_to_unicode(response),
    } for request, response in zip(
        cassette_dict['requests'],
        cassette_dict['responses'],
    )])
    data = {
        'version': CASSETTE_FORMAT_VERSION,
        'interactions': interactions,
        'recorded_with': 'VCR'
    }
    return serializer.serialize(data)


persist.serialize = github_serialize


class CivisVCRTestCase(VCRTestCase):

    def _get_vcr_kwargs(self):
        return {
            'filter_headers': ['Authorization'],
            'record_mode': 'once',
            'path_transformer': lambda x: x.replace('yaml', 'yml')
        }



import logging

from civis import APIClient
from civis._utils import maybe_get_random_name
from civis.futures import CivisFuture
from civis._deprecation import deprecate_param

log = logging.getLogger(__name__)


@deprecate_param('v2.0.0', 'api_key')
def query_civis(sql, database, api_key=None, client=None, credential_id=None,
                preview_rows=10, polling_interval=None, hidden=True):
    """Execute a SQL statement as a Civis query.

    Run a query that may return no results or where only a small
    preview is required. To execute a query that returns a large number
    of rows, see :func:`~civis.io.read_civis_sql`.

    Parameters
    ----------
    sql : str
        The SQL statement to execute.
    database : str or int
        The name or ID of the database.
    api_key : DEPRECATED str, optional
        Your Civis API key. If not given, the :envvar:`CIVIS_API_KEY`
        environment variable will be used.
    client : :class:`civis.APIClient`, optional
        If not provided, an :class:`civis.APIClient` object will be
        created from the :envvar:`CIVIS_API_KEY`.
    credential_id : str or int, optional
        The ID of the database credential. If ``None``, the default
        credential will be used.
    preview_rows : int, optional
        The maximum number of rows to return. No more than 100 rows can be
        returned at once.
    polling_interval : int or float, optional
        Number of seconds to wait between checks for query completion.
    hidden : bool, optional
        If ``True`` (the default), this job will not appear in the Civis UI.

    Returns
    -------
    results : :class:`~civis.futures.CivisFuture`
        A `CivisFuture` object.

    Examples
    --------
    >>> run = query_civis(sql="DELETE schema.table", database='database')
    >>> run.result()  # Wait for query to complete
    """
    if client is None:
        client = APIClient(api_key=api_key)
    database_id = client.get_database_id(database)
    cred_id = credential_id or client.default_credential
    resp = client.queries.post(database_id,
                               sql,
                               preview_rows,
                               credential=cred_id,
                               hidden=hidden)
    return CivisFuture(client.queries.get, (resp.id, ), polling_interval,
                       client=client, poll_on_creation=False)


@deprecate_param('v2.0.0', 'api_key')
def transfer_table(source_db, dest_db, source_table, dest_table,
                   job_name=None, api_key=None, client=None,
                   source_credential_id=None, dest_credential_id=None,
                   polling_interval=None, **advanced_options):
    """Transfer a table from one location to another.

    Parameters
    ----------
    source_db : str or int
        The name of the database where the source table is located.
        Optionally, could be the database ID.
    dest_db : str or int
        The name of the database where the table will be transfered.
        Optionally, could be the database ID.
    source_table : str
        Full name of the table to transfer, e.g., ``'schema.table'``.
    dest_table : str
        Full name of the table in the destination database, e.g.,
        ``'schema.table'``.
    job_name : str, optional
        A name to give the job. If omitted, a random job name will be
        used.
    api_key : DEPRECATED str, optional
        Your Civis API key. If not given, the :envvar:`CIVIS_API_KEY`
        environment variable will be used.
    client : :class:`civis.APIClient`, optional
        If not provided, an :class:`civis.APIClient` object will be
        created from the :envvar:`CIVIS_API_KEY`.
    source_credential_id : str or int, optional
        Optional credential ID for the source database. If ``None``, the
        default credential will be used.
    dest_credential_id : str or int, optional
        Optional credential ID for the destination database. If ``None``,
        the default credential will be used.
    polling_interval : int or float, optional
        Number of seconds to wait between checks for job completion.
    **advanced_options : kwargs
        Extra keyword arguments will be passed to the import sync job. See
        :func:`~civis.resources._resources.Imports.post_syncs`.

    Returns
    -------
    results : :class:`~civis.futures.CivisFuture`
        A `CivisFuture` object.

    Examples
    --------
    >>> transfer_table(source_db='Cluster A', dest_db='Cluster B',
    ...                source_table='schma.tbl', dest_table='schma.tbl')
    """
    if client is None:
        client = APIClient(api_key=api_key)
    source_cred_id = source_credential_id or client.default_credential
    dest_cred_id = dest_credential_id or client.default_credential
    job_name = maybe_get_random_name(job_name)
    source = {
        'remote_host_id': client.get_database_id(source_db),
        'credential_id': source_cred_id
    }
    destination = {
        'remote_host_id': client.get_database_id(dest_db),
        'credential_id': dest_cred_id
    }
    job_id = client.imports.post(job_name, "Dbsync", True, source=source,
                                 destination=destination).id

    client.imports.post_syncs(id=job_id,
                              source={'path': source_table},
                              destination={'path': dest_table},
                              advanced_options=advanced_options)
    run_id = client.imports.post_runs(id=job_id).run_id
    log.debug('Started run %d of sync for import %d', run_id, job_id)
    fut = CivisFuture(client.imports.get_files_runs, (job_id, run_id),
                      polling_interval=polling_interval, client=client,
                      poll_on_creation=False)
    return fut



import os
import re

import civis


_THIS_DIR = os.path.dirname(os.path.realpath(__file__))
_REPO_DIR = os.path.dirname(os.path.dirname(_THIS_DIR))


def test_version_number_match_with_changelog():
    """__version__ and CHANGELOG.md match for the latest version number."""
    changelog = open(os.path.join(_REPO_DIR, 'CHANGELOG.md')).read()
    version_in_changelog = (
        re.search(r'##\s+(\d+\.\d+\.\d+)', changelog).groups()[0])
    assert civis.__version__ == version_in_changelog, (
        'Make sure both __version__ and CHANGELOG are updated to match the '
        'latest version number')



"""Mock client creation and tooling
"""
from functools import lru_cache
import os
from unittest import mock

from civis import APIClient


TEST_SPEC = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                         "civis_api_spec.json")
TEST_SPEC_ALL = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                             "civis_api_spec_channels.json")


def create_client_mock(cache=TEST_SPEC):
    """Create an APIClient mock from a cache of the API spec

    Parameters
    ----------
    cache : str, optional
        Location of the API spec on the local filesystem

    Returns
    -------
    mock.Mock
        A `Mock` object which looks like an APIClient and which will
        error if any method calls have non-existent / misspelled parameters
    """
    # Create a client from the cache. We'll use this for auto-speccing.
    real_client = _real_client(cache)

    # Prevent the client from trying to talk to the real API when autospeccing
    with mock.patch('requests.Session', mock.MagicMock):
        mock_client = mock.create_autospec(real_client, spec_set=True)

    return mock_client


@lru_cache(maxsize=1)
def _real_client(local_api_spec):
    real_client = APIClient(local_api_spec=local_api_spec, api_key='none')
    real_client._feature_flags = {'noflag': None}
    if hasattr(real_client, 'channels'):
        # Deleting "channels" causes the client to fall back on
        # regular polling for completion, which greatly eases testing.
        delattr(real_client, 'channels')
    return real_client



from collections import OrderedDict
from functools import lru_cache
import json
import os
import re
import textwrap
try:
    from inspect import Signature, Parameter
except ImportError:
    from funcsigs import Signature, Parameter

from jsonref import JsonRef
import requests
from requests import Request

from civis.base import Endpoint, get_base_url
from civis._deprecation import deprecate_param
from civis._utils import (camel_to_snake, to_camelcase,
                          open_session, get_api_key,
                          retry_request, MAX_RETRIES)


API_VERSIONS = ["1.0"]
BASE_RESOURCES_V1 = [
    'aliases',
    'announcements',
    'apps',
    'civis',
    'clusters',
    'codes',
    'credentials',
    'databases',
    'endpoints',
    'enhancements',
    'exports',
    'files',
    'git_repos',
    'groups',
    'imports',
    'jobs',
    'json_values',
    'match_targets',
    'media',
    'models',
    'notebooks',
    'notifications',
    'ontology',
    'predictions',
    'projects',
    'queries',
    'remote_hosts',
    'reports',
    'results',
    'scripts',
    'search',
    'services',
    'storage_hosts',
    'tables',
    'templates',
    'users',
    'workflows'
]

TYPE_MAP = {"array": "list", "object": "dict"}
ITERATOR_PARAM_DESC = (
    "iterator : bool, optional\n"
    "    If True, return a generator to iterate over all responses. Use when\n"
    "    more results than the maximum allowed by limit are needed. When\n"
    "    True, limit and page_num are ignored. Defaults to False.\n")
CACHED_SPEC_PATH = os.path.join(os.path.expanduser('~'),
                                ".civis_api_spec.json")


@deprecate_param('v2.0.0', 'resources')
def exclude_resource(path, api_version, resources):
    if api_version == "1.0" and resources == "base":
        include = any([path.startswith(x) for x in BASE_RESOURCES_V1])
    else:
        include = True
    return not include


def get_properties(x):
    return x.get("properties") or x.get("items", {}).get("properties")


def property_type(props):
    t = TYPE_MAP.get(props['type'], props['type'])
    fmt = props.get("format")
    return "{}/{}".format(t, fmt) if fmt else t


def name_and_type_doc(name, prop, child, level, optional=False):
    """ Create a doc string element that includes a parameter's name
    and its type. This is intented to be combined with another
    doc string element that gives a description of the parameter.
    """
    prop_type = property_type(prop)
    snake_name = camel_to_snake(name)
    indent = " " * level * 4
    dash = "- " if level > 0 else ""
    colons = "::" if child else ""
    opt_str = ", optional" if optional else ""
    doc = "{}{}{} : {}{}{}"
    return doc.format(indent, dash, snake_name, prop_type, opt_str, colons)


def docs_from_property(name, prop, properties, level, optional=False):
    """ Create a list of doc string elements from a single property
    object. Avoids infinite recursion when a property contains a
    circular reference to its parent.
    """
    docs = []
    child_properties = get_properties(prop)
    child = None if child_properties == properties else child_properties
    docs.append(name_and_type_doc(name, prop, child, level, optional))
    doc_str = prop.get("description")
    if doc_str:
        indent = 4 * (level + 1) * " "
        doc_wrap = textwrap.fill(doc_str,
                                 initial_indent=indent,
                                 subsequent_indent=indent,
                                 width=79)
        docs.append(doc_wrap)
    if child:
        child_docs = docs_from_properties(child, level + 1)
        docs.append("\n".join(child_docs))
    return docs


def docs_from_properties(properties, level=0):
    """ Return doc string elements from a dictionary of properties."""
    docs = []
    for name, prop in properties.items():
        doc_list = docs_from_property(name, prop, properties, level)
        docs.extend(doc_list)
    return docs


def deprecated_notice(deprecation_warning):
    """ Return a doc string element for the deprecation notice. The
    doc string can be an empty string if the warning is None
    """
    if deprecation_warning is None:
        return ""

    return "Deprecation warning!\n------------------\n" + deprecation_warning


def doc_from_responses(responses):
    """ Return a doc string element from a responses object. The
    doc string describes the returned objects of a function.
    """
    response_code, response_object = next(iter(responses.items()))
    schema = response_object.get('schema', {})
    properties = get_properties(schema)
    if properties:
        result_doc = "\n".join(docs_from_properties(properties))
    else:
        description = response_object['description']
        result_doc_fmt = "None\n    Response code {}: {}"
        result_doc = result_doc_fmt.format(response_code, description)
    return "Returns\n-------\n" + result_doc


def join_doc_elements(*args):
    return "\n".join(args).rstrip()


def doc_from_param(param):
    """ Return a doc string element for a single parameter.
    Intended to be joined with other doc string elements to
    form a complete docstring of the accepted parameters of
    a function.
    """
    snake_name = camel_to_snake(param['name'])
    param_type = param['type']
    desc = param.get('description')
    optional = "" if param["required"] else ", optional"
    doc_body = ""
    if desc:
        indent = " " * 4
        doc_wrap = textwrap.fill(desc,
                                 initial_indent=indent,
                                 subsequent_indent=indent,
                                 width=79)
        doc_body += doc_wrap
        doc_body += "\n"
    doc_head = "{} : {}{}\n".format(snake_name, param_type, optional)
    return doc_head + doc_body


def iterable_method(method, params):
    """Determine whether it is possible for this endpoint to return an iterated
    response.
    """
    required_params = ('limit', 'page_num')
    params_present = all(param in params for param in required_params)
    return (method.lower() == 'get' and params_present)


def create_signature(args, optional_args):
    """ Dynamically create a signature for a function from strings.

    This function can be used to create a signature for a dynamically
    generated function without generating a string representation of
    the function code and making an explicit eval call.

    Parameters
    ----------
    args : list
        List of strings that name the required arguments of a function.
    optional_args : list
        List of strings that name the optional arguments of a function.

    Returns
    -------
    Signature(p) : inspect.Signature instance
        A Signature object that can be used to validate arguments in
        a dynamically created function.
    """
    p = [Parameter(x, Parameter.POSITIONAL_OR_KEYWORD) for x in args]
    p += [Parameter(x, Parameter.KEYWORD_ONLY, default='DEFAULT')
          for x in optional_args]
    return Signature(p)


def split_method_params(params):
    args = []
    optional_args = []
    body_params = []
    query_params = []
    path_params = []
    for param in params:
        name = param["name"]
        if param["required"]:
            args.append(name)
        else:
            optional_args.append(name)
        if param["in"] == "body":
            body_params.append(name)
        elif param["in"] == "query":
            query_params.append(name)
        elif param["in"] == "path":
            path_params.append(name)
    return args, optional_args, body_params, query_params, path_params


def create_method(params, verb, method_name, path, doc):
    """ Dynamically create a function to make an API call.

    The returned function accepts required parameters as positional arguments
    and optional parameters as kwargs.  The function passes these parameters
    into the appropriate place (path, query or body) in the API call,
    depending on the criteria of the API endpoint.

    Parameters
    ----------
    params : list
        A list of dictionaries that specify the arguments used
        by the returned function f
    verb : str
        HTTP verb to call
    method_name : str
        The name to give the returned function f
    path : str
        Endpoint path, possibly including replacement fields
        (i.e. scripts/{id})
    doc : str
        Documentation string for the returned function f


    Returns
    ------
    f : function
        A function which will make an API call
    """
    elements = split_method_params(params)
    sig_args, sig_opt_args, body_params, query_params, path_params = elements
    sig = create_signature(sig_args, sig_opt_args)
    is_iterable = iterable_method(verb, query_params)

    def f(self, *args, **kwargs):
        raise_for_unexpected_kwargs(method_name, kwargs, sig_args,
                                    sig_opt_args, is_iterable)

        iterator = kwargs.pop('iterator', False)
        arguments = sig.bind(*args, **kwargs).arguments
        if arguments.get("kwargs"):
            arguments.update(arguments.pop("kwargs"))
        body = {x: arguments[x] for x in body_params if x in arguments}
        query = {x: arguments[x] for x in query_params if x in arguments}
        path_vals = {x: arguments[x] for x in path_params if x in arguments}
        url = path.format(**path_vals) if path_vals else path
        return self._call_api(verb, url, query, body, iterator=iterator)

    # Add signature to function, including 'self' for class method
    sig_self = create_signature(["self"] + sig_args, sig_opt_args)
    f.__signature__ = sig_self
    f.__doc__ = doc
    f.__name__ = str(method_name)
    return f


def raise_for_unexpected_kwargs(method_name, arguments, sig_args, sig_kwargs,
                                is_iterable):
    """Raise TypeError if arguments are not in sig_args or sig_kwargs."""
    expected = set(sig_args) | set(sig_kwargs)
    if is_iterable:
        expected |= {'iterator'}
    unexpected = set(arguments.keys()) - expected
    if unexpected:
        msg_fmt = "{}() got an unexpected keyword argument(s) {}"
        raise TypeError(msg_fmt.format(method_name, str(unexpected)))


def bracketed(x):
    return re.search("^{.*}$", x)


def parse_param(param):
    """ Parse a parameter into a list of dictionaries which can
    be used to add the parameter to a dynamically generated function.
    """
    doc = ""
    args = []
    param_in = param['in']
    if param_in == 'body':
        body_args = parse_param_body(param)
        args.extend(body_args)
    else:
        snake_name = camel_to_snake(param['name'])
        req = param['required']
        doc = doc_from_param(param)
        a = {"name": snake_name, "in": param_in, "required": req, "doc": doc}
        args.append(a)
    return args


def parse_params(parameters, summary, verb):
    """ Parse the parameters of a function specification into a list
    of dictionaries which are used to generate the function at runtime.
    """
    args = []
    for param in parameters:
        args.extend(parse_param(param))
    if iterable_method(verb, (x["name"] for x in args)):
        iter_arg = {"name": "iterator", "in": None,
                    "required": False, "doc": ITERATOR_PARAM_DESC}
        args.append(iter_arg)
    req_docs = [x["doc"] for x in args if x["required"]]
    opt_docs = [x["doc"] for x in args if not x["required"]]
    param_docs = "".join(req_docs + opt_docs)
    if summary:
        summary_str = "{}\n".format(textwrap.fill(summary, width=79))
    else:
        summary_str = ""
    if param_docs:
        docs = "{}\nParameters\n----------\n{}".format(summary_str, param_docs)
    elif summary:
        docs = summary_str
    return args, docs


def parse_param_body(parameter):
    """ Parse the nested element of a parameter into a list of dictionaries
    which can be used to add the parameter to a dynamically generated
    function.
    """
    schema = parameter['schema']
    properties = schema['properties']
    req = schema.get('required', [])
    arguments = []
    for name, prop in properties.items():
        snake_name = camel_to_snake(name)
        is_req = name in req
        doc_list = docs_from_property(name, prop, properties, 0, not is_req)
        doc = "\n".join(doc_list) + "\n"
        a = {"name": snake_name, "in": "body", "required": is_req, "doc": doc}
        arguments.append(a)
    return arguments


def parse_method_name(verb, path):
    """ Create method name from endpoint path

    Create method name as the http verb (method) followed by
    any static path parameters. As there is ambiguity where
    a path may optionally have a path parameter passed (i.e.
    GET reports/ and GET reports/{id}), any GET method is changed
    to "list" for endpoints where the final path parameter is
    static.

    Examples
    --------
    >>> parse_method_name("get", "url.com/containers")
    list_containers
    >>> parse_method_name("get", "url.com/containers/{id}")
    get_containers
    >>> parse_method_name("get", "url.com/containers/{id}/runs/{run_id}")
    get_containers_runs
    >>> parse_method_name("post", "containers/{id}/runs/{run_id}")
    post_containers_runs
    """
    path_elems = path.split("/")[1:]
    name_elems = []
    for i, elem in enumerate(path_elems):
        prev_elem = path_elems[i - 1] if i > 0 else None
        if not bracketed(elem):
            name_elems.append(elem)
        elif prev_elem and bracketed(prev_elem):
            name_elems.append(prev_elem.strip('{|}'))
    final_elem = path_elems[-1] if path_elems else ""
    verb = "list" if verb == "get" and (not bracketed(final_elem)) else verb
    path_name = "_".join(name_elems)
    method_name = "_".join((verb, path_name)) if path_name else verb
    return re.sub("-", "_", method_name)


def parse_method(verb, operation, path):
    """ Generate a python function from a specification of that function."""
    summary = operation["summary"]
    params = operation.get("parameters", [])
    responses = operation["responses"]
    deprecation_warning = operation.get("x-deprecation-warning", None)
    if 'deprecated' in summary.lower():
        return None

    args, param_doc = parse_params(params, summary, verb)
    response_doc = doc_from_responses(responses)
    deprecation_notice = deprecated_notice(deprecation_warning)
    docs = join_doc_elements(deprecation_notice, param_doc, response_doc)
    name = parse_method_name(verb, path)

    method = create_method(args, verb, name, path, docs)
    return name, method


@deprecate_param('v2.0.0', 'resources')
def parse_path(path, operations, api_version, resources):
    """ Parse an endpoint into a class where each valid http request
    on that endpoint is converted into a convenience function and
    attached to the class as a method.
    """
    path = path.strip('/')
    modified_base_path = re.sub("-", "_", path.split('/')[0].lower())
    methods = []
    if exclude_resource(path, api_version, resources):
        return modified_base_path, methods
    for verb, op in operations.items():
        method = parse_method(verb, op, path)
        if method is None:
            continue
        methods.append(method)
    return modified_base_path, methods


@deprecate_param('v2.0.0', 'resources')
def parse_api_spec(api_spec, api_version, resources):
    """ Dynamically create classes to interface with the Civis API.

    Parse an OpenAPI (Swagger) specification into a dictionary of classes
    where each class represents an endpoint resource and contains
    methods to make http requests on that resource.

    Parameters
    ----------
    api_spec : OrderedDict
        The Civis API specification to parse.  References should be resolved
        before passing, typically using jsonref.JsonRef().
    api_version : string, optional
        The version of endpoints to call. May instantiate multiple client
        objects with different versions.  Currently only "1.0" is supported.
    resources : string, optional
        When set to "base", only the default endpoints will be exposed in the
        client object.  Set to "all" to include all endpoints available for
        a given user, including those that may be in development and subject
        to breaking changes at a later date.
    """
    paths = api_spec['paths']
    classes = {}
    for path, ops in paths.items():
        base_path, methods = parse_path(path, ops, api_version, resources)
        class_name = to_camelcase(base_path)
        if methods and classes.get(base_path) is None:
            classes[base_path] = type(str(class_name), (Endpoint,), {})
        for method_name, method in methods:
            setattr(classes[base_path], method_name, method)
    return classes


@lru_cache(maxsize=4)
def get_api_spec(api_key, api_version="1.0", user_agent="civis-python"):
    """Download the Civis API specification.

    Parameters
    ----------
    api_key : str
        Your API key obtained from the Civis Platform.
    api_version : string, optional
        The version of endpoints to call. May instantiate multiple client
        objects with different versions.  Currently only "1.0" is supported.
    user_agent : string, optional
        Provide this user agent to the the Civis API, along with an
        API client version tag and ``requests`` version tag.
    """
    if api_version == "1.0":
        with open_session(api_key, user_agent=user_agent) as sess:
            request = Request('GET', "{}endpoints".format(get_base_url()))
            pre_request = sess.prepare_request(request)
            response = retry_request('get', pre_request, sess, MAX_RETRIES)
    else:
        msg = "API specification for api version {} cannot be found"
        raise ValueError(msg.format(api_version))
    if response.status_code in (401, 403):
        msg = "{} error downloading API specification. API key may be expired."
        raise requests.exceptions.HTTPError(msg.format(response.status_code))
    response.raise_for_status()
    spec = response.json(object_pairs_hook=OrderedDict)
    return spec


@lru_cache(maxsize=4)
@deprecate_param('v2.0.0', 'resources')
def generate_classes(api_key, api_version="1.0", resources="all"):
    """ Dynamically create classes to interface with the Civis API.

    The Civis API documents behavior using an OpenAPI/Swagger specification.
    This function parses the specification and returns a dictionary of
    classes to provide a convenient method to make requests to the Civis
    API and to view documentation.

    https://github.com/OAI/OpenAPI-Specification

    Parameters
    ----------
    api_key : str
        Your API key obtained from the Civis Platform.
    api_version : string, optional
        The version of endpoints to call. May instantiate multiple client
        objects with different versions.  Currently only "1.0" is supported.
    resources : string, optional
        When set to "base", only the default endpoints will be exposed in the
        client object.  Set to "all" to include all endpoints available for
        a given user, including those that may be in development and subject
        to breaking changes at a later date.
    """
    assert api_version in API_VERSIONS, (
        "APIClient api_version must be one of {}".format(API_VERSIONS))
    assert resources in ["base", "all"], (
        "resources must be one of {}".format(["base", "all"]))
    raw_spec = get_api_spec(api_key, api_version)
    spec = JsonRef.replace_refs(raw_spec)
    return parse_api_spec(spec, api_version, resources)


def cache_api_spec(cache=CACHED_SPEC_PATH, api_key=None, api_version="1.0"):
    """Cache a local copy of the Civis Data Science API spec

    Parameters
    ----------
    cache : str, optional
        File in which to store the cache of the API spec
    api_key : str, optional
        Your API key obtained from the Civis Platform. If not given, the
        client will use the :envvar:`CIVIS_API_KEY` environment variable.
    api_version : string, optional
        The version of endpoints to call. May instantiate multiple client
        objects with different versions.  Currently only "1.0" is supported.
    """
    api_key = get_api_key(api_key)
    spec = get_api_spec(api_key, api_version=api_version)
    with open(cache, "wt") as _fout:
        json.dump(spec, _fout)


@deprecate_param('v2.0.0', 'resources')
def generate_classes_maybe_cached(cache, api_key, api_version, resources):
    """Generate class objects either from /endpoints or a local cache."""
    if cache is None:
        classes = generate_classes(api_key, api_version, resources)
    else:
        if isinstance(cache, OrderedDict):
            raw_spec = cache
        elif isinstance(cache, str):
            with open(cache, "r") as f:
                raw_spec = json.load(f, object_pairs_hook=OrderedDict)
        else:
            msg = "cache must be an OrderedDict or str, given {}"
            raise ValueError(msg.format(type(cache)))
        spec = JsonRef.replace_refs(raw_spec)
        classes = parse_api_spec(spec, api_version, resources)
    classes_ = _add_no_underscore_compatibility(classes)
    return classes_


def _add_no_underscore_compatibility(classes):
    """ Add class names without underscores for compatibility.

    Previously, no resources had underscores in APIClient.  Parsing of
    resources has been fixed so now these resources will have underscores.
    This adds an extra class for those resources without an underscore
    for compatibility. Additionally, this removes the resource "feature_flags"
    as APIClient has a name collision with this resource. This will
    be removed in v2.0.0.
    """
    new = ["bocce_clusters", "match_targets", "remote_hosts", "feature_flags"]
    classes_ = {}
    class_names = list(classes.keys())
    for class_name in class_names:
        if class_name != "feature_flags":
            classes_[class_name] = classes[class_name]
        if class_name in new:
            class_name_no_us = "".join(class_name.split("_"))
            classes_[class_name_no_us] = classes[class_name]
    return classes_



from collections import OrderedDict
import json
from unittest import mock

from civis import APIClient
from civis.resources._resources import get_api_spec, generate_classes
from civis.tests.testcase import CivisVCRTestCase
from civis.tests import TEST_SPEC

api_import_str = 'civis.resources._resources.get_api_spec'
with open(TEST_SPEC) as f:
    civis_api_spec = json.load(f, object_pairs_hook=OrderedDict)


class ClientTests(CivisVCRTestCase):

    @classmethod
    def setUpClass(cls):
        get_api_spec.cache_clear()
        generate_classes.cache_clear()

    @classmethod
    def tearDownClass(cls):
        get_api_spec.cache_clear()
        generate_classes.cache_clear()

    @mock.patch(api_import_str, return_value=civis_api_spec)
    def test_feature_flags(self, *mocks):
        client = APIClient()
        feature_flags = client.feature_flags
        expected = ('python_3_scripts', 'container_scripts', 'pubnub')
        self.assertCountEqual(feature_flags, expected)

    @mock.patch(api_import_str, return_value=civis_api_spec)
    def test_feature_flags_memoized(self, *mocks):
        client = APIClient()
        with mock.patch.object(client.users, 'list_me',
                               wraps=client.users.list_me):
            client.feature_flags
            client.feature_flags
            self.assertEqual(client.users.list_me.call_count, 1)



"""Tests for the test tooling
"""
import pytest

from civis.tests import mocks


def test_client_mock_attributeerror():
    mock_client = mocks.create_client_mock()
    with pytest.raises(AttributeError):
        mock_client.not_an_endpoint()


def test_client_mock_bad_parameter():
    mock_client = mocks.create_client_mock()
    mock_client.tables.list(database_id=1)  # Valid parameter
    with pytest.raises(TypeError):
        mock_client.tables.list(db_id=1)  # Invalid parameter



import os
import json
from collections import OrderedDict
from operator import itemgetter
from unittest import mock

import pytest

from civis import APIClient, response
from civis.base import CivisAPIError, CivisJobFailure
from civis.resources._resources import get_api_spec, generate_classes
from civis.futures import (ContainerFuture,
                           _ContainerShellExecutor,
                           CustomScriptExecutor,
                           _create_docker_command)

from civis.futures import (CivisFuture,
                           JobCompleteListener,
                           _LONG_POLLING_INTERVAL)
from civis.tests import TEST_SPEC, create_client_mock
from pubnub.enums import PNStatusCategory

from civis.tests.testcase import CivisVCRTestCase

api_import_str = 'civis.resources._resources.get_api_spec'
THIS_DIR = os.path.dirname(os.path.realpath(__file__))
with open(TEST_SPEC) as f:
    civis_api_spec_base = json.load(f, object_pairs_hook=OrderedDict)

with open(os.path.join(THIS_DIR, "civis_api_spec_channels.json")) as f:
    civis_api_spec_channels = json.load(f, object_pairs_hook=OrderedDict)


def clear_lru_cache():
    # LRU cache persists between tests so these caches need to be cleared
    # when different api specs are used in different test cases
    get_api_spec.cache_clear()
    generate_classes.cache_clear()


def setup_listener_status_mocks(status_category):
    match = mock.Mock()
    callback = mock.Mock()
    disconnect = mock.Mock()
    listener = JobCompleteListener(match, callback, disconnect)
    status = mock.Mock()
    status.category = status_category
    return match, callback, disconnect, listener, status


class CivisFutureTests(CivisVCRTestCase):

    @classmethod
    def setUpClass(cls):
        clear_lru_cache()

    @classmethod
    def tearDownClass(cls):
        clear_lru_cache()

    def test_listener_calls_callback_when_message_matches(self):
        match = mock.Mock()
        match.return_value = True
        callback = mock.Mock()
        listener = JobCompleteListener(match, callback)
        message = mock.Mock()
        message.message.return_value = 'test message'

        listener.message(None, message)
        match.assert_called_with(message.message)
        self.assertEqual(callback.call_count, 1)

    def test_listener_does_not_call_callback(self):
        match = mock.Mock()
        match.return_value = False
        callback = mock.Mock()
        listener = JobCompleteListener(match, callback)
        message = mock.Mock()
        message.message.return_value = 'test message'

        listener.message(None, message)
        match.assert_called_with(message.message)
        self.assertEqual(callback.call_count, 0)

    def test_listener_calls_disconnect_callback_when_status_disconnect(self):
        disconnect_categories = [
            PNStatusCategory.PNTimeoutCategory,
            PNStatusCategory.PNNetworkIssuesCategory,
            PNStatusCategory.PNUnexpectedDisconnectCategory,
        ]
        for category in disconnect_categories:
            mocks = setup_listener_status_mocks(category)
            _, _, disconnect, listener, status = mocks
            listener.status(None, status)
            assert disconnect.call_count == 1

    def test_listener_does_note_call_disconnect_callback_on_other_status(self):
        nondisconnect_categories = [
            PNStatusCategory.PNAcknowledgmentCategory,
            PNStatusCategory.PNConnectedCategory,
            PNStatusCategory.PNReconnectedCategory,
        ]
        for category in nondisconnect_categories:
            mocks = setup_listener_status_mocks(category)
            _, _, disconnect, listener, status = mocks
            listener.status(None, status)
            assert disconnect.call_count == 0

    @mock.patch(api_import_str, return_value=civis_api_spec_channels)
    @mock.patch.object(CivisFuture, '_subscribe')
    def test_check_message(self, *mocks):
        result = CivisFuture(lambda x: x, (1, 20))
        message = {
            'object': {
                'id': 1
            },
            'run': {
                'id': 20,
                'state': 'succeeded'
            }
        }
        self.assertTrue(result._check_message(message))

    @mock.patch(api_import_str, return_value=civis_api_spec_channels)
    @mock.patch.object(CivisFuture, '_subscribe')
    def test_check_message_with_different_run_id(self, *mocks):
        result = CivisFuture(lambda x: x, (1, 20))
        message = {
            'object': {
                'id': 2
            },
            'run': {
                'id': 20,
                'state': 'succeeded'
            }
        }
        self.assertFalse(result._check_message(message))

    @mock.patch(api_import_str, return_value=civis_api_spec_channels)
    @mock.patch.object(CivisFuture, '_subscribe')
    def test_check_message_when_job_is_running(self, *mocks):
        result = CivisFuture(lambda x: x, (1, 20))
        message = {
            'object': {
                'id': 1
            },
            'run': {
                'id': 20,
                'state': 'running'
            }
        }
        self.assertFalse(result._check_message(message))

    @mock.patch(api_import_str, return_value=civis_api_spec_channels)
    @mock.patch.object(CivisFuture, '_subscribe')
    def test_set_api_result_result_succeeded(self, mock_subscribe, mock_api):
        mock_pubnub = mock.Mock()
        mock_pubnub.unsubscribe_all.return_value = None
        mock_subscribe.return_value = mock_pubnub
        poller = mock.Mock()
        api_result = mock.Mock()
        api_result.state = 'succeeded'

        result = CivisFuture(poller, (1, 2))
        result._set_api_result(api_result)
        assert poller.call_count == 0
        assert mock_pubnub.unsubscribe_all.call_count == 1
        assert result._state == 'FINISHED'

    @mock.patch(api_import_str, return_value=civis_api_spec_channels)
    @mock.patch.object(CivisFuture, '_subscribe')
    def test_set_api_result_failed(self, mock_subscribe, mock_api):
        mock_pubnub = mock.Mock()
        mock_pubnub.unsubscribe_all.return_value = None
        mock_subscribe.return_value = mock_pubnub
        poller = mock.Mock()
        api_result = mock.Mock()
        api_result.state = 'failed'

        result = CivisFuture(poller, (1, 2))
        result._set_api_result(api_result)
        assert mock_pubnub.unsubscribe_all.call_count == 1
        assert result._state == 'FINISHED'
        with pytest.raises(CivisJobFailure):
            result.result()
        with pytest.raises(CivisJobFailure):
            result.outputs()

    def test_outputs_succeeded(self):
        poller = mock.Mock()
        api_result = mock.Mock()
        api_result.state = 'succeeded'
        mock_client = create_client_mock()
        expected_return = [{'test': 'test_result'}]
        mock_client.jobs.list_runs_outputs.return_value = expected_return

        result = CivisFuture(poller, (1, 2), client=mock_client)
        result._set_api_result(api_result)
        assert result.outputs() == expected_return

    @mock.patch(api_import_str, return_value=civis_api_spec_channels)
    @mock.patch.object(CivisFuture, '_subscribe')
    def test_subscribed_with_channels(self, *mocks):
        future = CivisFuture(lambda x: x,
                             (1, 20))
        future._pubnub.get_subscribed_channels.return_value = [1]
        assert future.subscribed is True

    @mock.patch(api_import_str, return_value=civis_api_spec_channels)
    @mock.patch.object(CivisFuture, '_subscribe')
    def test_subscribed_with_no_subscription(self, *mocks):
        future = CivisFuture(lambda x: x,
                             (1, 20))
        future._pubnub.get_subscribed_channels.return_value = []
        assert future.subscribed is False

    @mock.patch(api_import_str, return_value=civis_api_spec_base)
    @mock.patch.object(CivisFuture, '_subscribe')
    def test_subscribed_with_no_channels(self, *mocks):
        clear_lru_cache()
        future = CivisFuture(lambda x: x,
                             (1, 20))
        assert future.subscribed is False
        clear_lru_cache()

    @mock.patch(api_import_str, return_value=civis_api_spec_channels)
    @mock.patch.object(CivisFuture, '_subscribe')
    def test_overwrite_polling_interval_with_channels(self, *mocks):
        future = CivisFuture(lambda x: x, (1, 20))
        assert future.polling_interval == _LONG_POLLING_INTERVAL
        assert hasattr(future, '_pubnub')

    @mock.patch(api_import_str, return_value=civis_api_spec_channels)
    @mock.patch.object(CivisFuture, '_subscribe')
    def test_explicit_polling_interval_with_channels(self, *mocks):
        future = CivisFuture(lambda x: x, (1, 20), polling_interval=5)
        assert future.polling_interval == 5
        assert hasattr(future, '_pubnub')

    @mock.patch(api_import_str, return_value=civis_api_spec_base)
    @mock.patch.object(CivisFuture, '_subscribe')
    def test_polling_interval(self, *mocks):
        # This tests the fallback to polling when channels is not available.
        # It uses a different api spec than the other tests so it
        # should clear the cached values before and after
        clear_lru_cache()

        polling_interval = 30
        future = CivisFuture(lambda x: x,
                             (1, 20),
                             polling_interval=polling_interval)
        assert future.polling_interval == polling_interval
        assert hasattr(future, '_pubnub') is False

        clear_lru_cache()


def _check_executor(from_template_id=None):
    job_id, run_id = 42, 43
    c = _setup_client_mock(job_id, run_id, n_failures=0)
    mock_run = c.scripts.post_containers_runs()
    if from_template_id:
        bpe = CustomScriptExecutor(from_template_id=from_template_id,
                                   client=c, polling_interval=0.01)
        future = bpe.submit(my_param='spam')
    else:
        bpe = _ContainerShellExecutor(client=c, polling_interval=0.01)
        future = bpe.submit("foo")

    # Mock and test running, future.job_id, and done()
    mock_run.state = "running"
    assert future.running(), "future is incorrectly marked as not running"
    assert future.job_id == job_id, "job_id not stored properly"
    assert not future.done(), "future is incorrectly marked as done"

    future.cancel()

    # Mock and test cancelled()
    assert future.cancelled(), "cancelled() did not return True as expected"
    assert not future.running(), "running() did not return False as expected"

    # Mock and test done()
    mock_run.state = "succeeded"
    assert future.done(), "done() did not return True as expected"

    # Test cancelling all jobs.
    mock_run.state = "running"
    bpe.cancel_all()
    assert future.cancelled(), "cancel_all() failed"

    # Test shutdown method.
    bpe.shutdown(wait=True)
    assert future.done(), "shutdown() failed"

    return c


@pytest.mark.parametrize(
    'poller_args,expected_job_id,expected_run_id',
    [((123, 456), 123, 456),
     ((123,), 123, None)]
)
def test_future_job_id_run_id(poller_args, expected_job_id, expected_run_id):
    result = CivisFuture(
        poller=lambda x: x,
        poller_args=poller_args,
        client=create_client_mock(),
    )
    assert result.job_id == expected_job_id
    assert result.run_id == expected_run_id


def test_container_future_job_id_run_id():
    job_id, run_id = 123, 456
    result = ContainerFuture(
        job_id=job_id,
        run_id=run_id,
        client=create_client_mock(),
    )
    assert result.job_id == job_id
    assert result.run_id == run_id


def test_container_scripts():
    c = _check_executor()
    assert c.scripts.post_custom.call_count == 0
    assert c.scripts.post_containers.call_count > 0


def test_custom_scripts():
    with mock.patch.dict('os.environ', {'CIVIS_JOB_ID': '12',
                                        'CIVIS_RUN_ID': '40'}):
        c = _check_executor(133)
    assert c.scripts.post_custom.call_count > 0
    assert c.scripts.post_containers.call_count == 0

    # Verify that this script's job and run ID are passed to arguments
    args = c.scripts.post_custom.call_args[1].get('arguments')
    for k, v in (('CIVIS_PARENT_JOB_ID', '12'), ('CIVIS_PARENT_RUN_ID', '40')):
        assert args.get(k) == v


@pytest.mark.parametrize('is_child_job', [False, True])
def test_container_script_param_injection(is_child_job):
    # Test that child jobs created by the shell executor have the
    # job and run IDs of the script which created them (if any).
    job_id, run_id = '123', '13'
    c = _setup_client_mock(42, 43, n_failures=0)
    mock_env = {'CIVIS_JOB_ID': job_id, 'CIVIS_RUN_ID': run_id}

    with mock.patch.dict('os.environ', mock_env):
        init_kwargs = dict(client=c, polling_interval=0.01)
        if is_child_job:
            init_kwargs['params'] = [
                {'name': 'CIVIS_PARENT_JOB_ID', 'type': 'integer',
                 'value': '888'},
                {'name': 'CIVIS_PARENT_RUN_ID', 'type': 'integer',
                 'value': '999'},
            ]
        bpe = _ContainerShellExecutor(**init_kwargs)
        bpe.submit("foo")

    params = sorted(c.scripts.post_containers.call_args[1].get('params'),
                    key=itemgetter('name'))
    assert params == [
        {'name': 'CIVIS_PARENT_JOB_ID', 'type': 'integer', 'value': job_id},
        {'name': 'CIVIS_PARENT_RUN_ID', 'type': 'integer', 'value': run_id}
    ], "The parent job parameters were not set correctly."


def test_create_docker_command():
    res = _create_docker_command("foo.sh", "bar", "baz", wibble="wibble1",
                                 wobble="wobble1")
    assert res == "foo.sh bar baz --wibble wibble1 --wobble wobble1"


# A function to raise fake API errors the first
# num_failures times it is called.
def _make_error_func(num_failures, failure_is_error=False):
    """Raise API errors multiple times before succeeding

    Test error-handling code by using this to mock
    calls to post_containers_runs or get_containers_runs.

    Parameters
    ----------
    num_failures: int
        Fail this many times before returning a success
    failure_is_error: bool
        If True, "failure" means raising a `CivisAPIError`.

    Returns
    -------
    MockRun
        Mock which imitates the result of a `post_containers_runs`
        or `get_containers_runs` call
    """
    counter = {'failures': 0}  # Use a dict so we can modify it in the closure

    def mock_api_error(job_id, run_id):
        if counter['failures'] < num_failures:
            counter['failures'] += 1
            if failure_is_error:
                raise CivisAPIError(mock.MagicMock())
            else:
                return response.Response({'id': run_id,
                                          'container_id': job_id,
                                          'state': 'failed'})
        else:
            return response.Response({'id': run_id,
                                      'container_id': job_id,
                                      'state': 'succeeded'})
    return mock_api_error


def _setup_client_mock(job_id=-10, run_id=100, n_failures=8,
                       failure_is_error=False):
    """Return a Mock set up for use in testing container scripts

    Parameters
    ----------
    job_id: int
        Mock-create containers with this ID when calling `post_containers`
        or `post_containers_runs`.
    run_id: int
        Mock-create runs with this ID when calling `post_containers_runs`.
    n_failures: int
        When calling `get_containers_runs`, fail this many times
        before succeeding.
    failure_is_error: bool
        If True, "failure" means raising a `CivisAPIError`.

    Returns
    -------
    `unittest.mock.Mock`
        With `post_containers`, `post_containers_runs`, and
        `get_containers_runs` methods set up.
    """
    c = mock.Mock()
    c.__class__ = APIClient

    mock_container = response.Response({'id': job_id})
    c.scripts.post_containers.return_value = mock_container
    c.scripts.post_custom.return_value = mock_container
    mock_container_run = response.Response({'id': run_id,
                                            'container_id': job_id,
                                            'state': 'queued'})
    c.scripts.post_containers_runs.return_value = mock_container_run
    c.jobs.post_runs.return_value = mock_container_run
    c.scripts.get_containers_runs.side_effect = _make_error_func(
        n_failures, failure_is_error)

    def change_state_to_cancelled(job_id):
        mock_container_run.state = "cancelled"
        return mock_container_run

    c.scripts.post_cancel.side_effect = change_state_to_cancelled
    del c.channels  # Remove "channels" endpoint to fall back on polling

    return c


def test_cancel_finished_job():
    # If we try to cancel a completed job, we get a 404 error.
    # That shouldn't be sent to the user.

    # Set up a mock client which will give an exception when
    # you try to cancel any job.
    c = _setup_client_mock()
    err_resp = response.Response({
        'status_code': 404,
        'error': 'not_found',
        'errorDescription': 'The requested resource could not be found.',
        'content': True})
    err_resp.json = lambda: err_resp.json_data
    c.scripts.post_cancel.side_effect = CivisAPIError(err_resp)
    c.scripts.post_containers_runs.return_value.state = 'running'

    fut = ContainerFuture(-10, 100, polling_interval=1, client=c,
                          poll_on_creation=False)
    assert not fut.done()
    assert fut.cancel() is False


def test_future_no_retry_error():
    # Verify that with no retries, exceptions on job polling
    #  are raised to the user
    c = _setup_client_mock(failure_is_error=True)
    fut = ContainerFuture(-10, 100, polling_interval=0.001, client=c)
    with pytest.raises(CivisAPIError):
        fut.result()


def test_future_no_retry_failure():
    # Verify that with no retries, job failures are raised as
    # exceptions for the user
    c = _setup_client_mock(failure_is_error=False)
    fut = ContainerFuture(-10, 100, polling_interval=0.001, client=c)
    with pytest.raises(CivisJobFailure):
        fut.result()


def test_future_not_enough_retry_error():
    # Verify that if polling the run is still erroring after all retries
    # are exhausted, the error will be raised for the user.
    c = _setup_client_mock(failure_is_error=True)
    fut = ContainerFuture(-10, 100, max_n_retries=3, polling_interval=0.01,
                          client=c)
    with pytest.raises(CivisAPIError):
        fut.result()


def test_future_not_enough_retry_failure():
    # Verify that if the job is still failing after all retries
    # are exhausted, the job failure will be raised for the user.
    c = _setup_client_mock(failure_is_error=False)
    fut = ContainerFuture(-10, 100, max_n_retries=3, polling_interval=0.01,
                          client=c)
    with pytest.raises(CivisJobFailure):
        fut.result()


def test_future_retry_failure():
    # Verify that we can retry through API errors until a job succeeds
    c = _setup_client_mock(failure_is_error=False)
    fut = ContainerFuture(-10, 100, max_n_retries=10, polling_interval=0.01,
                          client=c)
    assert fut.result().state == 'succeeded'


def test_future_retry_error():
    # Verify that we can retry through job failures until it succeeds
    c = _setup_client_mock(failure_is_error=True)
    fut = ContainerFuture(-10, 100, max_n_retries=10, polling_interval=0.01,
                          client=c)
    assert fut.result().state == 'succeeded'



"""Run CivisML jobs and retrieve the results
"""
import builtins
from builtins import super
import collections
from functools import lru_cache
import io
import json
import logging
import os
import re
import shutil
import tempfile
import threading
import warnings
from concurrent import futures
from functools import wraps

import joblib
try:
    from sklearn.base import BaseEstimator
    HAS_SKLEARN = True
except ImportError:
    HAS_SKLEARN = False

from civis import APIClient, find, find_one
from civis._utils import camel_to_snake
from civis.base import CivisAPIError, CivisJobFailure
import civis.io as cio
from civis.futures import ContainerFuture
from civis.response import Response


__all__ = ['ModelFuture', 'ModelError', 'ModelPipeline']
log = logging.getLogger(__name__)

# sentinel value for default primary key value
SENTINEL = collections.namedtuple('Sentinel', [])()


class ModelError(RuntimeError):
    def __init__(self, msg, estimator=None, metadata=None):
        self.metadata = metadata
        self.estimator = estimator
        super().__init__(msg)


def _check_fit_initiated(method):
    """Makes sure that the ModelPipeline's been trained"""
    @wraps(method)
    def wrapper(*args, **kwargs):
        self = args[0]
        if not self.train_result_:
            raise ValueError("This model hasn't been trained yet.")
        return method(*args, **kwargs)
    return wrapper


def _block_and_handle_missing(method):
    """For ModelFuture file-retrieving property methods.
    Block until completion and attempt to retrieve result.
    Raise exception only if the result isn't found.
    """
    @wraps(method)
    def wrapper(self):
        futures.wait((self,))  # Block until done
        try:
            return method(self)
        except FileNotFoundError:
            # We get here if the modeling job failed to produce
            # any output and we don't have metadata.
            if self.exception():
                raise self.exception() from None
            else:
                raise
    return wrapper


def _stash_local_dataframe(df, template_id, client=None):
    """Store data in a temporary Civis File and return the file ID"""
    # Standard dataframe indexes do not have a "levels" attribute,
    # but multiindexes do. Checking for this attribute means we don't
    # need to import pandas to do error handling here.
    if getattr(getattr(df, "index", None), "levels", None) is not None:
        raise TypeError("CivisML does not support multi-indexed data frames. "
                        "Try calling `.reset_index` on your data to convert "
                        "it into a CivisML-friendly format.")
    try:
        if template_id > 9969:
            return _stash_dataframe_as_feather(df, client)
        else:
            return _stash_dataframe_as_csv(df, client)
    except (ImportError, AttributeError) as exc:
        if (df.dtypes == 'category').any():
            # The original exception should tell users if they need
            # to upgrade pandas (an AttributeError)
            # # or if they need to install "feather-format" (ImportError).
            raise ValueError(
                'Categorical columns can only be handled with pandas '
                'version >= 0.20 and `feather-format` installed.') from exc
        return _stash_dataframe_as_csv(df, client)


def _stash_dataframe_as_feather(df, client):
    civis_fname = 'modelpipeline_data.feather'
    with tempfile.TemporaryDirectory() as tdir:
        path = os.path.join(tdir, civis_fname)
        df.to_feather(path)
        file_id = cio.file_to_civis(path, name=civis_fname, client=client)
    return file_id


def _stash_dataframe_as_csv(df, client):
    civis_fname = 'modelpipeline_data.csv'
    txt = io.StringIO()
    df.to_csv(txt, encoding='utf-8', index=False)
    txt.flush()
    txt.seek(0)
    file_id = cio.file_to_civis(txt, name=civis_fname, client=client)

    return file_id


def _stash_local_file(csv_path, client=None):
    """Store data in a temporary Civis File and return the file ID"""
    civis_fname = 'modelpipeline_data.csv'
    with open(csv_path) as _fin:
        file_id = cio.file_to_civis(_fin, name=civis_fname, client=client)

    return file_id


def _decode_train_run(train_job_id, train_run_id, client):
    """Determine correct run ID for use for a given training job ID"""
    try:
        return int(train_run_id)
    except ValueError:
        container = client.scripts.get_containers(int(train_job_id))
        if train_run_id == 'active':
            train_run_id = container.arguments.get('ACTIVE_BUILD', find_one(
                container.params, name='ACTIVE_BUILD'))['default']

        if train_run_id == 'latest':
            return container.last_run.id

        try:
            return int(train_run_id)
        except Exception as exc:
            msg = ('Please provide valid train_run_id! Needs to be '
                   'integer corresponding to a training run ID '
                   'or one of "active" or "latest".')
            raise ValueError(msg) from exc


def _retrieve_file(fname, job_id, run_id, local_dir, client=None):
    """Download a Civis file using a reference on a previous run"""
    file_id = cio.file_id_from_run_output(fname, job_id, run_id, client=client)
    fpath = os.path.join(local_dir, fname)
    # fname may contain a path
    output_dir = os.path.dirname(fpath)
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
    with open(fpath, 'wb') as down_file:
        cio.civis_to_file(file_id, down_file, client=client)
    return fpath


def _load_table_from_outputs(job_id, run_id, filename, client=None,
                             **table_kwargs):
    """Load a table from a run output directly into a ``DataFrame``"""
    client = APIClient() if client is None else client
    file_id = cio.file_id_from_run_output(filename, job_id, run_id,
                                          client=client, regex=True)
    return cio.file_to_dataframe(file_id, client=client, **table_kwargs)


def _load_estimator(job_id, run_id, filename='estimator.pkl', client=None):
    """Load a joblib-serialized Estimator from run outputs"""
    try:
        tempdir = tempfile.mkdtemp()
        path = _retrieve_file(filename, job_id, run_id, tempdir, client=client)
        obj = joblib.load(path)
    finally:
        shutil.rmtree(tempdir)
    return obj


def _exception_from_logs(exc, job_id, run_id, client, nlog=15):
    """Create an exception if the log has a recognizable error

    Search "error" emits in the last ``n_log`` lines.
    This function presently recognizes the following errors:

    - MemoryError
    """
    logs = client.scripts.list_containers_runs_logs(job_id, run_id, limit=nlog)

    # Check for memory errors
    msgs = [x['message'] for x in logs if x['level'] == 'error']
    mem_err = [m for m in msgs if m.startswith('Process ran out of its')]
    if mem_err:
        exc = MemoryError(mem_err[0])
    else:
        # Unknown error; return logs to the user as a sort of traceback
        all_logs = '\n'.join([x['message'] for x in logs])
        if isinstance(exc, CivisJobFailure):
            exc.error_message = all_logs + '\n' + exc.error_message
        else:
            exc = CivisJobFailure(all_logs)
    return exc


def _parse_warning(warn_str):
    """Reverse-engineer a warning string

    Parameters
    ----------
    warn_str : string

    Returns
    -------
    (str, Warning, str, int)
        message, category, filename, lineno
    """
    tokens = warn_str.rstrip('\n').split(" ")

    # The first token is
    # "[filename]:[lineno]:"
    filename, lineno, _ = tokens[0].split(':')

    # The second token is
    # "[category name]:"
    category = getattr(builtins, tokens[1][:-1], RuntimeWarning)

    message = " ".join(tokens[2:])

    return message, category, filename, int(lineno)


def _show_civisml_warnings(warn_list):
    """Re-raise warnings recorded during a CivisML run

    Parameters
    ----------
    warn_list : list of str
        A list of warnings generated during a CivisML run
    """
    for warn_str in warn_list:
        try:
            warnings.warn_explicit(*_parse_warning(warn_str))
        except Exception:  # NOQA
            warn_str = "Remote warning from CivisML:\n" + warn_str
            warnings.warn(warn_str, RuntimeWarning)


def _get_job_type_version(alias):
    """Derive the job type and version from the given alias.

    Parameters
    ----------
    alias : str
        CivisML alias

    Returns
    -------
    str
        Job type, one of {training, prediction, registration}.
    str
        CivisML version, e.g., "v2.2".
    """
    # A version-less alias for production, e.g., "civis-civisml-training"
    match_production = re.search(r'\Acivis-civisml-(\w+)\Z', alias)
    # A versioned alias, e.g., "civis-civisml-training-v2-3"
    match_v = re.search(r'\Acivis-civisml-(\w+)-v(\d+)-(\d+)\Z', alias)
    # A special-version alias, e.g., "civis-civisml-training-dev"
    match_special = re.search(r'\Acivis-civisml-(\w+)-(\S+[^-])\Z', alias)

    if match_production:
        job_type = match_production.group(1)
        version = None
    elif match_v:
        job_type = match_v.group(1)
        version = 'v%s.%s' % match_v.group(2, 3)
    elif match_special:
        job_type = match_special.group(1)
        version = match_special.group(2)
    else:
        msg = ('Unable to parse the job type and version '
               'from the CivisML alias "%r"')
        raise ValueError(msg % alias)

    return job_type, version


@lru_cache()
def _get_template_ids_all_versions(client):
    """Get templates IDs for all accessible CivisML versions.

    Parameters
    ----------
    client : APIClient
        Civis API client object

    Returns
    -------
    Dict[str, Dict[str, int]]
        Mapping between versions (e.g., "v2.2") and template IDs for the given
        version (e.g., {'training': 1, 'prediction': 2, 'registration': 3}).
    """
    template_alias_objects = client.aliases.list(
        object_type='template_script', iterator=True
    )
    civisml_template_alias_objects = find(
        template_alias_objects,
        alias=lambda alias: alias.startswith('civis-civisml-')
    )
    ids = collections.defaultdict(
        lambda: {'training': None, 'prediction': None, 'registration': None}
    )
    for alias_obj in civisml_template_alias_objects:
        try:
            job_type, version = _get_job_type_version(alias_obj.alias)
        except ValueError:
            msg = (
                '%r looks like a CivisML alias for the prefix "civis-civisml-"'
                ', but it is impossible to parse its job type and version'
            )
            log.debug(msg % alias_obj)
            continue
        ids[version][job_type] = alias_obj.object_id
    if not ids:
        r = Response({'status_code': 404,
                      'reason': 'No CivisML template IDs are accessible.',
                      'content': None})
        raise CivisAPIError(r)
    # Disallow a defaultdict in the output, so that a non-existent CivisML
    # version as key should trigger a KeyError.
    ids = dict(ids)
    return ids


def _get_template_ids(civisml_version, client):
    """Get template IDs for the specified CivisML version.

    Parameters
    ----------
    civisml_version : str
        CivisML version
    client : APIClient
        Civis API client object

    Returns
    -------
    int
        Template ID for training
    int
        Template ID for prediction
    int
        Template ID for pre-trained model registration
    """
    template_ids_all_versions = _get_template_ids_all_versions(client)
    try:
        ids = template_ids_all_versions[civisml_version]
    except KeyError:
        msg = (
            '"{civisml_version}" is an invalid CivisML version. '
            'Either this version does not exist, or you do not have access '
            'to this version. '
            'Versions accessible to you are {{{accessible_versions}}}, '
            'as well as `None` for the latest production version.'
        ).format(
            civisml_version=civisml_version,
            accessible_versions=', '.join(
                '"%s"' % v
                # Don't include None, or else it would crash sorted()
                for v in sorted(
                    v for v in template_ids_all_versions.keys() if v
                )
            )
        )
        raise ValueError(msg)
    return ids['training'], ids['prediction'], ids['registration']


class ModelFuture(ContainerFuture):
    """Encapsulates asynchronous execution of a CivisML job

    This object knows where to find modeling outputs
    from CivisML jobs. All data attributes are
    lazily retrieved and block on job completion.

    This object can be pickled, but it does not store the state
    of the attached :class:`~civis.APIClient` object. An unpickled
    ModelFuture will use the API key from the user's environment.

    Parameters
    ----------
    job_id : int
        ID of the modeling job
    run_id : int
        ID of the modeling run
    train_job_id : int, optional
        If not provided, this object is assumed to encapsulate a training
        job, and ``train_job_id`` will equal ``job_id``.
    train_run_id : int, optional
        If not provided, this object is assumed to encapsulate a training
        run, and ``train_run_id`` will equal ``run_id``.
    polling_interval : int or float, optional
        The number of seconds between API requests to check whether a result
        is ready. The default intelligently switches between a short
        interval if ``pubnub`` is not available and a long interval
        for ``pubnub`` backup if that library is installed.
    client : :class:`civis.APIClient`, optional
        If not provided, an :class:`civis.APIClient` object will be
        created from the :envvar:`CIVIS_API_KEY`.
    poll_on_creation : bool, optional
        If ``True`` (the default), it will poll upon calling ``result()`` the
        first time. If ``False``, it will wait the number of seconds specified
        in `polling_interval` from object creation before polling.

    Attributes
    ----------
    metadata : dict, blocking
        The metadata associated with this modeling job
    metrics : dict, blocking
        Validation metrics from this job's training
    validation_metadata : dict, blocking
        Metadata from this modeling job's validation run
    train_metadata : dict, blocking
        Metadata from this modeling job's training run
        (will be identical to `metadata` if this is a training run)
    estimator : :class:`sklearn.pipeline.Pipeline`, blocking
        The fitted scikit-learn Pipeline resulting from this model run
    table : :class:`pandas.DataFrame`, blocking
        The table output from this modeling job: out-of-sample
        predictions on the training set for a training job, or
        a table of predictions for a prediction job.
        If the prediction job was split into multiple files
        (this happens automatically for large tables),
        this attribute will provide only predictions for the first file.
    state : str
        The current state of the Civis Platform run
    job_id : int
    run_id : int
    train_job_id : int
        Container ID for the training job -- identical to ``job_id``
        if this is a training job.
    train_run_id : int
        As ``train_job_id`` but for runs
    is_training : bool
        True if this ``ModelFuture`` corresponds to a train-validate job.

    Methods
    -------
    cancel()
        Cancels the corresponding Platform job before completion
    succeeded()
        (Non-blocking) Is the job a success?
    failed()
        (Non-blocking) Did the job fail?
    cancelled()
        (Non-blocking) Was the job cancelled?
    running()
        (Non-blocking) Is the job still running?
    done()
        (Non-blocking) Is the job finished?
    result()
        (Blocking) Return the final status of the Civis Platform job.

    See Also
    --------
    civis.futures.CivisFuture
    civis.futures.ContainerFuture
    concurrent.futures.Future
    """

    def __init__(self, job_id, run_id, train_job_id=None, train_run_id=None,
                 polling_interval=None, client=None, poll_on_creation=True):
        super().__init__(job_id, run_id,
                         polling_interval=polling_interval,
                         client=client,
                         poll_on_creation=poll_on_creation)
        if train_job_id and train_run_id:
            self.is_training = False
            self.train_job_id = train_job_id
            self.train_run_id = train_run_id
        else:
            self.is_training = True
            self.train_job_id = self.job_id
            self.train_run_id = self.run_id
        self._metadata, self._val_metadata = None, None
        self._train_data, self._train_data_fname = None, None
        self._train_metadata = None
        self._table, self._estimator = None, None
        self._exception_handled = False
        self.add_done_callback(self._set_model_exception)

    @staticmethod
    def _set_model_exception(fut):
        """Callback: On job completion, check the metadata.
        If it indicates an exception, replace the generic
        ``CivisJobFailure`` by a more informative ``ModelError``.
        """
        # Prevent infinite recursion: this function calls `set_exception`,
        # which triggers callbacks (i.e. re-calls this function).
        if fut._exception_handled:
            return
        else:
            fut._exception_handled = True

        try:
            meta = fut.metadata
            if fut.is_training and meta['run']['status'] == 'succeeded':
                # if training job and job succeeded, check validation job
                meta = fut.validation_metadata
            if meta is not None and meta['run']['status'] == 'exception':
                try:
                    # This will fail if the user doesn't have joblib installed
                    est = fut.estimator
                except Exception:  # NOQA
                    est = None
                fut.set_exception(
                    ModelError('Model run failed with stack trace:\n'
                               '{}'.format(meta['run']['stack_trace']),
                               est, meta))
        except (FileNotFoundError, CivisJobFailure) as exc:
            # If there's no metadata file
            # (we get FileNotFound or CivisJobFailure),
            # check the tail of the log for a clearer exception.
            exc = _exception_from_logs(exc, fut.job_id, fut.run_id, fut.client)
            fut.set_exception(exc)
        except futures.CancelledError:
            # We don't need to change the exception if the run was cancelled
            pass
        except KeyError:
            # KeyErrors always represent a bug in the modeling code,
            # but showing the resulting KeyError can be confusing and
            # mask the real error.
            warnings.warn("Received malformed metadata from Civis Platform. "
                          "Something went wrong with job execution.")

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['_polling_thread']
        del state['client']
        del state['poller']
        del state['_condition']
        if '_pubnub' in state:
            state['_pubnub'] = True  # Replace with a boolean flag
        state['_done_callbacks'] = []
        state['_self_polling_executor'] = None

        return state

    def __setstate__(self, state):
        self.__dict__ = state
        self._condition = threading.Condition()
        self.client = APIClient()
        self.poller = self.client.scripts.get_containers_runs
        self._begin_tracking()
        self.add_done_callback(self._set_model_exception)

    @property
    def state(self):
        state = self._civis_state
        if state == 'succeeded':
            state = self.metadata['run']['status']
        return state

    @property
    def table_fname(self):
        return 'predictions.csv'

    @property
    def metadata_fname(self):
        return 'model_info.json'

    @property
    def train_data_fname(self):
        if self._train_data_fname is None:
            self._train_data_fname = os.path.basename(self.training_metadata
                                                      .get('run')
                                                      .get('configuration')
                                                      .get('data')
                                                      .get('location'))
        return self._train_data_fname

    @property
    def train_data(self):
        if self._train_data is None:
            try:
                self._train_data = _load_table_from_outputs(
                    self.train_job_id,
                    self.train_run_id,
                    self.train_data_fname,
                    client=self.client)
            except CivisAPIError as err:
                if err.status_code == 404:
                    msg = 'There is no training data stored for this job!'
                    raise ValueError(msg) from err
                else:
                    raise

        return self._train_data

    def _table_primary_key(self):
        # metadata path to input parameters is different
        # for training and prediction
        if self.is_training:
            pkey = self.metadata[
                'run']['configuration']['data']['primary_key']
        else:
            pkey = self.metadata[
                'jobs'][0]['run']['configuration']['data']['primary_key']
        return pkey

    @property
    def table(self):
        self.result()  # Block and raise errors if any
        if self._table is None:
            # An index column will only be present if primary key is
            if self._table_primary_key() is None:
                index_col = False
            else:
                index_col = 0

            if self.is_training:
                try:
                    # Training jobs only have one output table, the OOS scores
                    self._table = _load_table_from_outputs(self.job_id,
                                                           self.run_id,
                                                           self.table_fname,
                                                           index_col=index_col,
                                                           client=self.client)
                except FileNotFoundError:
                    # Just pass here, because we want the table to stay None
                    # if it does not exist
                    pass
            else:
                # Prediction jobs may have many output tables.
                output_ids = self.metadata['output_file_ids']
                if len(output_ids) > 1:
                    print('This job output {} files. Retrieving only the '
                          'first. Find the full list at `metadata'
                          '["output_file_ids"]`.'.format(len(output_ids)))
                self._table = cio.file_to_dataframe(output_ids[0],
                                                    client=self.client,
                                                    index_col=index_col)
        return self._table

    @property
    @_block_and_handle_missing
    def metadata(self):
        if self._metadata is None:
            fid = cio.file_id_from_run_output('model_info.json', self.job_id,
                                              self.run_id, client=self.client)
            self._metadata = cio.file_to_json(fid, client=self.client)
            _show_civisml_warnings(self._metadata.get('warnings', []))
        return self._metadata

    @property
    @_block_and_handle_missing
    def estimator(self):
        if self._estimator is None:
            self._estimator = _load_estimator(self.train_job_id,
                                              self.train_run_id,
                                              client=self.client)
        return self._estimator

    @property
    @_block_and_handle_missing
    def validation_metadata(self):
        if self._val_metadata is None:
            try:
                fid = cio.file_id_from_run_output('metrics.json',
                                                  self.train_job_id,
                                                  self.train_run_id,
                                                  client=self.client)
            except FileNotFoundError:
                # Use an empty dictionary to indicate that
                # we've already checked for metadata.
                self._val_metadata = {}
            else:
                self._val_metadata = cio.file_to_json(fid, client=self.client)
        if not self._val_metadata:
            # Convert an empty dictionary to None
            return None
        else:
            return self._val_metadata

    @property
    def metrics(self):
        if self.validation_metadata:
            return self.validation_metadata['metrics']
        else:
            return None

    @property
    @_block_and_handle_missing
    def training_metadata(self):
        if self._train_metadata is None:
            fid = cio.file_id_from_run_output('model_info.json',
                                              self.train_job_id,
                                              self.train_run_id,
                                              client=self.client)
            self._train_metadata = cio.file_to_json(fid, client=self.client)
        return self._train_metadata


class ModelPipeline:
    """Interface for scikit-learn modeling in the Civis Platform

    Each ModelPipeline corresponds to a scikit-learn
    :class:`~sklearn.pipeline.Pipeline` which will run in Civis Platform.

    Note that this object can be safely pickled and unpickled, but it
    does not store the state of any attached :class:`~civis.APIClient` object.
    An unpickled ModelPipeline will use the API key from the user's
    environment.

    Parameters
    ----------
    model : string or Estimator
        Either the name of a pre-defined model
        (e.g. "sparse_logistic" or "gradient_boosting_classifier")
        or else a pre-existing Estimator object.
    dependent_variable : string or List[str]
        The dependent variable of the training dataset.
        For a multi-target problem, this should be a list of
        column names of dependent variables. Nulls in a single
        dependent variable will automatically be dropped.
    primary_key : string, optional
        The unique ID (primary key) of the training dataset.
        This will be used to index the out-of-sample scores.
    parameters : dict, optional
        Specify parameters for the final stage estimator in a
        predefined model, e.g. ``{'C': 2}`` for a "sparse_logistic"
        model.
    cross_validation_parameters : dict or string, optional
        Options for cross validation. For grid search, supply a
        parameter grid as a dictionary, e.g.,
        ``{{'n_estimators': [100, 200, 500], 'learning_rate': [0.01, 0.1],
        'max_depth': [2, 3]}}``. For hyperband, pass the string "hyperband".
    model_name : string, optional
        The prefix of the Platform modeling jobs. It will have
        " Train" or " Predict" added to become the Script title.
    calibration : {None, "sigmoid", "isotonic"}
        If not None, calibrate output probabilities with the selected method.
        Valid only with classification models.
    excluded_columns : array, optional
        A list of columns which will be considered ineligible to be
        independent variables.
    client : :class:`~civis.APIClient`, optional
        If not provided, an :class:`~civis.APIClient` object will be
        created from the :envvar:`CIVIS_API_KEY`.
    cpu_requested : int, optional
        Number of CPU shares requested in the Civis Platform for
        training jobs. 1024 shares = 1 CPU.
    memory_requested : int, optional
        Memory requested from Civis Platform for training jobs, in MiB
    disk_requested : float, optional
        Disk space requested on Civis Platform for training jobs, in GB
    notifications : dict
        See :func:`~civis.resources._resources.Scripts.post_custom` for
        further documentation about email and URL notification.
    dependencies : array, optional
        List of packages to install from PyPI or git repository (e.g., Github
        or Bitbucket). If a private repo is specified, please include a
        ``git_token_name`` argument as well (see below). Make sure to pin
        dependencies to a specific version, since dependencies will be
        reinstalled during every training and predict job.
    git_token_name : str, optional
        Name of remote git API token stored in Civis Platform as the password
        field in a custom platform credential. Used only when installing
        private git repositories.
    verbose : bool, optional
        If True, supply debug outputs in Platform logs and make
        prediction child jobs visible.
    etl : Estimator, optional
        Custom ETL estimator which overrides the default ETL, and
        is run before training and validation.
    civisml_version : str, optional
        CivisML version to use for training and prediction.
        If not provided, the latest version in production is used.

    Methods
    -------
    train()
        Train the model on data in Civis Platform; outputs
        :class:`~civis.ml.ModelFuture`
    predict()
        Make predictions on new data; outputs :class:`~civis.ml.ModelFuture`
    from_existing()
        Class method; use to create a :class:`~civis.ml.ModelPipeline`
        from an existing model training run

    Attributes
    ----------
    estimator : :class:`~sklearn.pipeline.Pipeline`
        The trained scikit-learn Pipeline
    train_result_ : :class:`~civis.ml.ModelFuture`
        :class:`~civis.ml.ModelFuture` encapsulating this model's training run
    state : str
        Status of the training job (non-blocking)

    Examples
    --------
    >>> from civis.ml import ModelPipeline
    >>> model = ModelPipeline('gradient_boosting_classifier', 'depvar',
    ...                       primary_key='voterbase_id')
    >>> train = model.train(table_name='schema.survey_data',
    ...                     fit_params={'sample_weight': 'survey_weight'},
    ...                     database_name='My Redshift Cluster',
    ...                     oos_scores='scratch.survey_depvar_oos_scores')
    >>> train
    <ModelFuture at 0x11be7ae10 state=queued>
    >>> train.running()
    True
    >>> train.done()
    False
    >>> df = train.table  # Read OOS scores from its Civis File. Blocking.
    >>> meta = train.metadata  # Metadata from training run
    >>> train.metrics['roc_auc']
    0.88425
    >>> pred = model.predict(table_name='schema.demographics_table ',
    ...                      database_name='My Redshift Cluster',
    ...                      output_table='schema.predicted_survey_response',
    ...                      if_exists='drop')
    >>> df_pred = pred.table  # Blocks until finished
    # Modify the parameters of the base estimator in a default model:
    >>> model = ModelPipeline('sparse_logistic', 'depvar',
    ...                       primary_key='voterbase_id',
    ...                       parameters={'C': 2})
    # Grid search over hyperparameters in the base estimator:
    >>> model = ModelPipeline('sparse_logistic', 'depvar',
    ...                       primary_key='voterbase_id',
    ...                       cross_validation_parameters={'C': [0.1, 1, 10]})

    See Also
    --------
    civis.ml.ModelFuture
    """
    def __init__(self, model, dependent_variable,
                 primary_key=None, parameters=None,
                 cross_validation_parameters=None, model_name=None,
                 calibration=None, excluded_columns=None, client=None,
                 cpu_requested=None, memory_requested=None,
                 disk_requested=None, notifications=None,
                 dependencies=None, git_token_name=None, verbose=False,
                 etl=None, civisml_version=None):
        self.model = model
        self._input_model = model  # In case we need to modify the input
        if isinstance(dependent_variable, str):
            # Standardize the dependent variable as a list.
            dependent_variable = [dependent_variable]
        self.dependent_variable = dependent_variable

        # optional but common parameters
        self.primary_key = primary_key
        self.parameters = parameters or {}
        self.cv_params = cross_validation_parameters or {}
        self.model_name = model_name  # None lets Platform use template name
        self.excluded_columns = excluded_columns
        self.calibration = calibration
        self.job_resources = {'REQUIRED_CPU': cpu_requested,
                              'REQUIRED_MEMORY': memory_requested,
                              'REQUIRED_DISK_SPACE': disk_requested}
        self.notifications = notifications or {}
        self.dependencies = dependencies
        self.git_token_name = git_token_name
        self.verbose = verbose

        if client is None:
            client = APIClient()
        self._client = client
        self.train_result_ = None

        template_ids = _get_template_ids(civisml_version, self._client)
        self.train_template_id, self.predict_template_id, _ = template_ids

        self.etl = etl
        if self.train_template_id < 9968 and self.etl is not None:
            # This is a pre-v2.0 CivisML template
            raise NotImplementedError("The etl argument is not implemented"
                                      " in this version of CivisML.")

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['_client']
        return state

    def __setstate__(self, state):
        self.__dict__ = state
        self._client = APIClient()

    @classmethod
    def register_pretrained_model(cls, model, dependent_variable=None,
                                  features=None, primary_key=None,
                                  model_name=None, dependencies=None,
                                  git_token_name=None,
                                  skip_model_check=False, verbose=False,
                                  client=None, civisml_version=None):
        """Use a fitted scikit-learn model with CivisML scoring

        Use this function to set up your own fitted scikit-learn-compatible
        Estimator object for scoring with CivisML. This function will
        upload your model to Civis Platform and store enough metadata
        about it that you can subsequently use it with a CivisML scoring job.

        The only required input is the model itself, but you are strongly
        recommended to also provide a list of feature names. Without a list
        of feature names, CivisML will have to assume that your scoring
        table contains only the features needed for scoring (perhaps also
        with a primary key column), in all in the correct order.

        Parameters
        ----------
        model : sklearn.base.BaseEstimator or int
            The model object. This must be a fitted scikit-learn compatible
            Estimator object, or else the integer Civis File ID of a
            pickle or joblib-serialized file which stores such an object.
            If an Estimator object is provided, it will be uploaded to the
            Civis Files endpoint and set to be available indefinitely.
        dependent_variable : string or List[str], optional
            The dependent variable of the training dataset.
            For a multi-target problem, this should be a list of
            column names of dependent variables.
        features : string or List[str], optional
            A list of column names of features which were used for training.
            These will be used to ensure that tables input for prediction
            have the correct features in the correct order.
        primary_key : string, optional
            The unique ID (primary key) of the scoring dataset
        model_name : string, optional
            The name of the Platform registration job. It will have
            " Predict" added to become the Script title for predictions.
        dependencies : array, optional
            List of packages to install from PyPI or git repository (e.g.,
            GitHub or Bitbucket). If a private repo is specified, please
            include a ``git_token_name`` argument as well (see below).
            Make sure to pin dependencies to a specific version, since
            dependencies will be reinstalled during every predict job.
        git_token_name : str, optional
            Name of remote git API token stored in Civis Platform as
            the password field in a custom platform credential.
            Used only when installing private git repositories.
        skip_model_check : bool, optional
            If you're sure that your model will work with CivisML, but it
            will fail the comprehensive verification, set this to True.
        verbose : bool, optional
            If True, supply debug outputs in Platform logs and make
            prediction child jobs visible.
        client : :class:`~civis.APIClient`, optional
            If not provided, an :class:`~civis.APIClient` object will be
            created from the :envvar:`CIVIS_API_KEY`.
        civisml_version : str, optional
            CivisML version to use.
            If not provided, the latest version in production is used.

        Returns
        -------
        :class:`~civis.ml.ModelPipeline`

        Examples
        --------
        This example assumes that you already have training data
        ``X`` and ``y``, where ``X`` is a :class:`~pandas.DataFrame`.

        >>> from civis.ml import ModelPipeline
        >>> from sklearn.linear_model import Lasso
        >>> est = Lasso().fit(X, y)
        >>> model = ModelPipeline.register_pretrained_model(
        ...     est, 'concrete', features=X.columns)
        >>> model.predict(table_name='my.table', database_name='my-db')
        """
        client = client or APIClient()

        if isinstance(dependent_variable, str):
            dependent_variable = [dependent_variable]
        if isinstance(features, str):
            features = [features]
        if isinstance(dependencies, str):
            dependencies = [dependencies]
        if not model_name:
            model_name = ("Pretrained {} model for "
                          "CivisML".format(model.__class__.__name__))
            model_name = model_name[:255]  # Max size is 255 characters

        if isinstance(model, (int, float, str)):
            model_file_id = int(model)
        else:
            try:
                tempdir = tempfile.mkdtemp()
                fout = os.path.join(tempdir, 'model_for_civisml.pkl')
                joblib.dump(model, fout, compress=3)
                with open(fout, 'rb') as _fout:
                    # NB: Using the name "estimator.pkl" means that
                    # CivisML doesn't need to copy this input to a file
                    # with a different name.
                    model_file_id = cio.file_to_civis(
                        _fout, 'estimator.pkl', expires_at=None, client=client)
            finally:
                shutil.rmtree(tempdir)

        args = {'MODEL_FILE_ID': str(model_file_id),
                'SKIP_MODEL_CHECK': skip_model_check,
                'DEBUG': verbose}
        if dependent_variable is not None:
            args['TARGET_COLUMN'] = ' '.join(dependent_variable)
        if features is not None:
            args['FEATURE_COLUMNS'] = ' '.join(features)
        if dependencies is not None:
            args['DEPENDENCIES'] = ' '.join(dependencies)
        if git_token_name:
            creds = find(client.credentials.list(),
                         name=git_token_name,
                         type='Custom')
            if len(creds) > 1:
                raise ValueError("Unique credential with name '{}' for "
                                 "remote git hosting service not found!"
                                 .format(git_token_name))
            args['GIT_CRED'] = creds[0].id

        _, _, template_id = _get_template_ids(civisml_version, client)
        if template_id is None:
            msg = (
                'No registration template ID is available. '
                'Pre-trained model registration is available for CivisML '
                'v2.2 (for which `civisml_version` would be "v2.2") or above, '
                'but you have specified CivisML version "%r"'
            )
            raise ValueError(msg % civisml_version)
        container = client.scripts.post_custom(
            from_template_id=template_id,
            name=model_name,
            arguments=args)
        log.info('Created custom script %s.', container.id)

        run = client.scripts.post_custom_runs(container.id)
        log.debug('Started job %s, run %s.', container.id, run.id)

        fut = ModelFuture(container.id, run.id, client=client,
                          poll_on_creation=False)
        fut.result()
        log.info('Model registration complete.')

        mp = ModelPipeline.from_existing(fut.job_id, fut.run_id, client)
        mp.primary_key = primary_key
        return mp

    @classmethod
    def from_existing(cls, train_job_id, train_run_id='latest', client=None):
        """Create a :class:`ModelPipeline` object from existing model IDs

        Parameters
        ----------
        train_job_id : int
            The ID of the CivisML job in the Civis Platform
        train_run_id : int or string, optional
            Location of the model run, either

            * an explicit run ID,
            * "latest" : The most recent run
            * "active" : The run designated by the training job's
              "active build" parameter
        client : :class:`~civis.APIClient`, optional
            If not provided, an :class:`~civis.APIClient` object will be
            created from the :envvar:`CIVIS_API_KEY`.

        Returns
        -------
        :class:`~civis.ml.ModelPipeline`
            A :class:`~civis.ml.ModelPipeline` which refers to
            a previously-trained model

        Examples
        --------
        >>> from civis.ml import ModelPipeline
        >>> model = ModelPipeline.from_existing(job_id)
        >>> model.train_result_.metrics['roc_auc']
        0.843
        """
        train_job_id = int(train_job_id)  # Convert np.int to int
        if client is None:
            client = APIClient()
        train_run_id = _decode_train_run(train_job_id, train_run_id, client)
        try:
            fut = ModelFuture(train_job_id, train_run_id, client=client)
            container = client.scripts.get_containers(train_job_id)
        except CivisAPIError as api_err:
            if api_err.status_code == 404:
                msg = ('There is no Civis Platform job with '
                       'script ID {} and run ID {}!'.format(train_job_id,
                                                            train_run_id))
                raise ValueError(msg) from api_err
            raise

        args = container.arguments

        # Older templates used "WORKFLOW" instead of "MODEL"
        model = args.get('MODEL', args.get('WORKFLOW'))
        dependent_variable = args['TARGET_COLUMN'].split()
        primary_key = args.get('PRIMARY_KEY')
        parameters = json.loads(args.get('PARAMS', "{}"))
        cross_validation_parameters = json.loads(args.get('CVPARAMS', "{}"))
        calibration = args.get('CALIBRATION')
        excluded_columns = args.get('EXCLUDE_COLS', None)
        if excluded_columns:
            excluded_columns = excluded_columns.split()
        cpu_requested = args.get('REQUIRED_CPU')
        memory_requested = args.get('REQUIRED_MEMORY')
        disk_requested = args.get('REQUIRED_DISK_SPACE')
        name = container.name
        if name.endswith(' Train'):
            # Strip object-applied suffix
            name = name[:-len(' Train')]
        notifications = {camel_to_snake(key): val for key, val
                         in container.notifications.items()}
        dependencies = args.get('DEPENDENCIES', None)
        if dependencies:
            dependencies = dependencies.split()
        git_token_name = args.get('GIT_CRED', None)
        if git_token_name:
            git_token_name = client.credentials.get(git_token_name).name

        klass = cls(model=model,
                    dependent_variable=dependent_variable,
                    primary_key=primary_key,
                    model_name=name,
                    parameters=parameters,
                    cross_validation_parameters=cross_validation_parameters,
                    calibration=calibration,
                    excluded_columns=excluded_columns,
                    client=client,
                    cpu_requested=cpu_requested,
                    disk_requested=disk_requested,
                    memory_requested=memory_requested,
                    notifications=notifications,
                    dependencies=dependencies,
                    git_token_name=git_token_name,
                    verbose=args.get('DEBUG', False))
        klass.train_result_ = fut

        # Set prediction template corresponding to training
        # or registration template
        template_id = int(container['from_template_id'])
        ids = find_one(
            _get_template_ids_all_versions(client).values(),
            lambda ids: ids['training'] == template_id or ids['registration'] == template_id  # noqa
        )
        p_id = ids['prediction']
        klass.predict_template_id = p_id

        return klass

    def train(self, df=None, csv_path=None, table_name=None,
              database_name=None, file_id=None,
              sql_where=None, sql_limit=None, oos_scores=None,
              oos_scores_db=None, if_exists='fail', fit_params=None,
              polling_interval=None, validation_data='train', n_jobs=None):
        """Start a Civis Platform job to train your model

        Provide input through one of
        a :class:`~pandas.DataFrame` (``df``),
        a local CSV (``csv_path``),
        a Civis Table (``table_name`` and ``database_name``), or
        a Civis File containing a CSV (``file_id``).

        Model outputs will always contain out-of-sample scores
        (accessible through :attr:`ModelFuture.table` on this function's
        output), and you may chose to store these out-of-sample scores
        in a Civis Table with the ``oos_scores``, ``oos_scores_db``,
        and ``if_exists`` parameters.

        Parameters
        ----------
        df : pd.DataFrame, optional
            A :class:`~pandas.DataFrame` of training data.
            The :class:`~pandas.DataFrame` will be uploaded to a Civis file so
            that CivisML can access it.
            Note that the index of the :class:`~pandas.DataFrame` will be
            ignored -- use ``df.reset_index()`` if you want your
            index column to be included with the data passed to CivisML.
            NB: You must install ``feather-format`` if your
            :class:`~pandas.DataFrame` contains :class:`~pandas.Categorical`
            columns, to ensure that CivisML preserves data types.
        csv_path : str, optional
            The location of a CSV of data on the local disk.
            It will be uploaded to a Civis file.
        table_name : str, optional
            The qualified name of the table containing the training set from
            which to build the model.
        database_name : str, optional
            Name of the database holding the training set table used to
            build the model. E.g., 'My Cluster Name'.
        file_id : int, optional
            If the training data are stored in a Civis file,
            provide the integer file ID.
        sql_where : str, optional
            A SQL WHERE clause used to scope the rows of the training set
            (used for table input only)
        sql_limit : int, optional
            SQL LIMIT clause for querying the training set
            (used for table input only)
        oos_scores : str, optional
            If provided, store out-of-sample predictions on
            training set data to this Redshift "schema.tablename".
        oos_scores_db : str, optional
            If not provided, store OOS predictions in the same database
            which holds the training data.
        if_exists : {'fail', 'append', 'drop', 'truncate'}
            Action to take if the out-of-sample prediction table
            already exists.
        fit_params: Dict[str, str]
            Mapping from parameter names in the model's ``fit`` method
            to the column names which hold the data, e.g.
            ``{'sample_weight': 'survey_weight_column'}``.
        polling_interval : float, optional
            Check for job completion every this number of seconds.
            Do not set if using the notifications endpoint.
        validation_data : str, optional
            Source for validation data. There are currently two options:
            `'train'` (the default), which cross-validates over training data
            for validation; and `'skip'`, which skips the validation step.
        n_jobs : int, optional
            Number of jobs to use for training and validation. Defaults to
            `None`, which allows CivisML to dynamically calculate an
            appropriate number of workers to use (in general, as many as
            possible without using all resources in the cluster).
            Increase n_jobs to parallelize over many hyperparameter
            combinations in grid search/hyperband, or decrease to use fewer
            computational resources at once.

        Returns
        -------
        :class:`~civis.ml.ModelFuture`
        """
        if ((table_name is None or database_name is None) and
                file_id is None and df is None and csv_path is None):
            raise ValueError('Provide a source of data.')
        if sum((bool(table_name and database_name),
                bool(file_id), df is not None, csv_path is not None)) > 1:
            raise ValueError('Provide a single source of data.')

        if df is not None:
            file_id = _stash_local_dataframe(df, self.train_template_id,
                                             client=self._client)
        elif csv_path:
            file_id = _stash_local_file(csv_path, client=self._client)

        train_args = {'TARGET_COLUMN': ' '.join(self.dependent_variable),
                      'PRIMARY_KEY': self.primary_key,
                      'PARAMS': json.dumps(self.parameters),
                      'CVPARAMS': json.dumps(self.cv_params),
                      'CALIBRATION': self.calibration,
                      'IF_EXISTS': if_exists}
        if oos_scores:
            train_args['OOSTABLE'] = oos_scores
        if oos_scores_db:
            oos_db_id = self._client.get_database_id(oos_scores_db)
            train_args['OOSDB'] = {'database': oos_db_id}
        if sql_where:
            train_args['WHERESQL'] = sql_where
        if sql_limit:
            train_args['LIMITSQL'] = sql_limit
        if self.excluded_columns:
            train_args['EXCLUDE_COLS'] = ' '.join(self.excluded_columns)
        if fit_params:
            train_args['FIT_PARAMS'] = json.dumps(fit_params)
        if self.dependencies:
            train_args['DEPENDENCIES'] = ' '.join(self.dependencies)
        if self.train_template_id >= 9968:
            if validation_data:
                train_args['VALIDATION_DATA'] = validation_data
            if n_jobs:
                train_args['N_JOBS'] = n_jobs

        if HAS_SKLEARN and isinstance(self.model, BaseEstimator):
            try:
                tempdir = tempfile.mkdtemp()
                fout = os.path.join(tempdir, 'estimator.pkl')
                joblib.dump(self.model, fout, compress=3)
                with open(fout, 'rb') as _fout:
                    n = self.model_name if self.model_name else "CivisML"
                    estimator_file_id = cio.file_to_civis(
                        _fout, 'Estimator for ' + n, client=self._client)
                self._input_model = self.model  # Keep the estimator
                self.model = str(estimator_file_id)
            finally:
                shutil.rmtree(tempdir)
        train_args['MODEL'] = self.model

        if HAS_SKLEARN and self.train_template_id >= 9968:
            if isinstance(self.etl, BaseEstimator):
                try:
                    tempdir = tempfile.mkdtemp()
                    fout = os.path.join(tempdir, 'ETL.pkl')
                    joblib.dump(self.etl, fout, compress=3)
                    with open(fout, 'rb') as _fout:
                        etl_file_id = cio.file_to_civis(
                            _fout, 'ETL Estimator', client=self._client)
                    train_args['ETL'] = str(etl_file_id)
                finally:
                    shutil.rmtree(tempdir)

        name = self.model_name + ' Train' if self.model_name else None
        # Clear the existing training result so we can make a new one.
        self.train_result_ = None

        result, container, run = self._create_custom_run(
            self.train_template_id,
            job_name=name,
            table_name=table_name,
            database_name=database_name,
            file_id=file_id,
            args=train_args,
            resources=self.job_resources,
            polling_interval=polling_interval)

        self.train_result_ = result

        return result

    def _create_custom_run(self, template_id, job_name=None, table_name=None,
                           database_name=None, file_id=None, args=None,
                           resources=None, polling_interval=None):
        # Handle int-like but non-Python-integer types such as np.int64
        file_id = int(file_id) if file_id is not None else file_id
        script_arguments = {'TABLE_NAME': table_name,
                            'CIVIS_FILE_ID': file_id,
                            'DEBUG': self.verbose}
        if database_name:
            if template_id < 8000:
                # v0 jobs used a different database parameter
                script_arguments['DB_NAME'] = database_name
            else:
                db_id = self._client.get_database_id(database_name)
                script_arguments['DB'] = {'database': db_id}
        resources = resources or {}
        for key, value in resources.items():
            if value:
                # Default resources are set on the template. Only
                # modify via arguments if users give a non-default value.
                script_arguments[key] = value
        if self.git_token_name:
            creds = find(self._client.credentials.list(),
                         name=self.git_token_name,
                         type='Custom')
            if len(creds) > 1:
                raise ValueError("Unique credential with name '{}' for "
                                 "remote git hosting service not found!"
                                 .format(self.git_token_name))
            script_arguments['GIT_CRED'] = creds[0].id

        script_arguments.update(args or {})

        container = self._client.scripts.post_custom(
            from_template_id=template_id,
            name=job_name,
            arguments=script_arguments,
            notifications=self.notifications)
        log.info('Created custom script %s.', container.id)

        run = self._client.scripts.post_custom_runs(container.id)
        log.debug('Started job %s, run %s.', container.id, run.id)

        train_kwargs = {}
        if self.train_result_ is not None:
            train_kwargs = {'train_job_id': self.train_result_.job_id,
                            'train_run_id': self.train_result_.run_id}
        fut = ModelFuture(
            container.id,
            run.id,
            client=self._client,
            polling_interval=polling_interval,
            poll_on_creation=False,
            **train_kwargs)

        return fut, container, run

    @property
    @_check_fit_initiated
    def state(self):
        return self.train_result_.state

    @property
    @_check_fit_initiated
    def estimator(self):
        return self.train_result_.estimator

    @_check_fit_initiated
    def predict(self, df=None, csv_path=None,
                table_name=None, database_name=None,
                manifest=None, file_id=None, sql_where=None, sql_limit=None,
                primary_key=SENTINEL, output_table=None, output_db=None,
                if_exists='fail', n_jobs=None, polling_interval=None,
                cpu=None, memory=None, disk_space=None,
                dvs_to_predict=None):
        """Make predictions on a trained model

        Provide input through one of
        a :class:`~pandas.DataFrame` (``df``),
        a local CSV (``csv_path``),
        a Civis Table (``table_name`` and ``database_name``),
        a Civis File containing a CSV (``file_id``), or
        a Civis File containing a manifest file (``manifest``).

        A "manifest file" is JSON which specifies the location of
        many shards of the data to be used for prediction.
        A manifest file is the output of a Civis
        export job with ``force_multifile=True`` set,
        e.g. from :func:`civis.io.civis_to_multifile_csv`.
        Large Civis Tables (provided using ``table_name``)
        will automatically be exported to manifest files.

        Prediction outputs will always be stored as gzipped
        CSVs in one or more Civis Files. You can find a list of
        File ID numbers for output files at the "output_file_ids"
        key in the metadata returned by the prediction job.
        Provide an ``output_table`` (and optionally an ``output_db``,
        if it's different from ``database_name``) to copy these
        predictions into a Civis Table.

        Parameters
        ----------
        df : pd.DataFrame, optional
            A :class:`~pandas.DataFrame` of data for prediction.
            The :class:`~pandas.DataFrame` will be uploaded to a Civis file so
            that CivisML can access it.
            Note that the index of the :class:`~pandas.DataFrame` will be
            ignored -- use ``df.reset_index()`` if you want your
            index column to be included with the data passed to CivisML.
            NB: You must install ``feather-format`` if your
            :class:`~pandas.DataFrame` contains :class:`~pandas.Categorical`
            columns, to ensure that CivisML preserves data types.
        csv_path : str, optional
            The location of a CSV of data on the local disk.
            It will be uploaded to a Civis file.
        table_name : str, optional
            The qualified name of the table containing your data
        database_name : str, optional
            Name of the database holding the
            data, e.g., 'My Redshift Cluster'.
        manifest : int, optional
            ID for a manifest file stored as a Civis file.
            (Note: if the manifest is not a Civis Platform-specific manifest,
            like the one returned from :func:`civis.io.civis_to_multfile_csv`,
            this must be used in conjunction with table_name and database_name
            due to the need for column discovery via Redshift.)
        file_id : int, optional
            If the data are a CSV stored in a Civis file,
            provide the integer file ID.
        sql_where : str, optional
            A SQL WHERE clause used to scope the rows to be predicted
        sql_limit : int, optional
            SQL LIMIT clause to restrict the size of the prediction set
        primary_key : str, optional
            Primary key of the prediction table. Defaults to
            the primary key of the training data. Use ``None`` to
            indicate that the prediction data don't have a
            primary key column.
        output_table: str, optional
            The table in which to put the predictions.
        output_db : str, optional
            Database of the output table. Defaults to the database
            of the input table.
        if_exists : {'fail', 'append', 'drop', 'truncate'}
            Action to take if the prediction table already exists.
        n_jobs : int, optional
            Number of concurrent Platform jobs to use
            for multi-file / large table prediction. Defaults to
            `None`, which allows CivisML to dynamically calculate an
            appropriate number of workers to use (in general, as many as
            possible without using all resources in the cluster).
        polling_interval : float, optional
            Check for job completion every this number of seconds.
            Do not set if using the notifications endpoint.
        cpu : int, optional
            CPU shares requested by the user for a single job.
        memory : int, optional
            RAM requested by the user for a single job.
        disk_space : float, optional
            disk space requested by the user for a single job.
        dvs_to_predict : list of str, optional
            If this is a multi-output model, you may list a subset of
            dependent variables for which you wish to generate predictions.
            This list must be a subset of the original `dependent_variable`
            input. The scores for the returned subset will be identical to
            the scores which those outputs would have had if all outputs
            were written, but ignoring some of the model's outputs will
            let predictions complete faster and use less disk space.
            The default is to produce scores for all DVs.

        Returns
        -------
        :class:`~civis.ml.ModelFuture`
        """
        self.train_result_.result()  # Blocks and raises training errors

        if ((table_name is None or database_name is None) and
                file_id is None and df is None and csv_path is None and
                manifest is None):
            raise ValueError('Provide a source of data.')
        if sum((bool(table_name and database_name) or (manifest is not None),
                bool(file_id), df is not None, csv_path is not None)) > 1:
            raise ValueError('Provide a single source of data.')

        if df is not None:
            file_id = _stash_local_dataframe(df, self.predict_template_id,
                                             client=self._client)
        elif csv_path:
            file_id = _stash_local_file(csv_path, client=self._client)

        if primary_key is SENTINEL:
            primary_key = self.primary_key

        predict_args = {'TRAIN_JOB': self.train_result_.job_id,
                        'TRAIN_RUN': self.train_result_.run_id,
                        'PRIMARY_KEY': primary_key,
                        'IF_EXISTS': if_exists}
        if output_table:
            predict_args['OUTPUT_TABLE'] = output_table
        if output_db:
            if self.predict_template_id == 7021:
                # v0 jobs used a different database parameter
                predict_args['OUTPUT_DB'] = output_db
            else:
                output_db_id = self._client.get_database_id(output_db)
                predict_args['OUTPUT_DB'] = {'database': output_db_id}

        if manifest:
            predict_args['MANIFEST'] = manifest
        if sql_where:
            predict_args['WHERESQL'] = sql_where
        if sql_limit:
            predict_args['LIMITSQL'] = sql_limit
        if n_jobs:
            predict_args['N_JOBS'] = n_jobs
        if dvs_to_predict:
            if isinstance(dvs_to_predict, str):
                dvs_to_predict = [dvs_to_predict]
            if self.predict_template_id > 10583:
                # This feature was added in v2.2; 10583 is the v2.1 template
                predict_args['TARGET_COLUMN'] = ' '.join(dvs_to_predict)
        if self.predict_template_id >= 9969:
            if cpu:
                predict_args['CPU'] = cpu
            if memory:
                predict_args['MEMORY'] = memory
            if disk_space:
                predict_args['DISK_SPACE'] = disk_space

        name = self.model_name + ' Predict' if self.model_name else None
        result, container, run = self._create_custom_run(
            self.predict_template_id,
            job_name=name,
            table_name=table_name,
            database_name=database_name,
            file_id=file_id,
            args=predict_args,
            polling_interval=polling_interval)

        return result



import logging

from civis import APIClient
from civis.futures import CivisFuture
from civis._deprecation import deprecate_param

log = logging.getLogger(__name__)


@deprecate_param("v2.0.0", "api_key")
def run_job(job_id, api_key=None, client=None, polling_interval=None):
    """Run a job.

    Parameters
    ----------
    job_id: str or int
        The ID of the job.
    api_key: DEPRECATED str, optional
        Your Civis API key. If not given, the :envvar:`CIVIS_API_KEY`
        environment variable will be used.
    client: :class:`civis.APIClient`, optional
        If not provided, an :class:`civis.APIClient` object will be
        created from the :envvar:`CIVIS_API_KEY`.
    polling_interval : int or float, optional
        The number of seconds between API requests to check whether a result
        is ready.

    Returns
    -------
    results: :class:`~civis.futures.CivisFuture`
        A `CivisFuture` object.
    """
    if client is None:
        client = APIClient(api_key=api_key)
    run = client.jobs.post_runs(job_id)
    return CivisFuture(
        client.jobs.get_runs,
        (job_id, run["id"]),
        client=client,
        polling_interval=polling_interval,
        poll_on_creation=False,
    )


def run_template(id, arguments, JSONValue=False, client=None):
    """Run a template and return the results.

    Parameters
    ----------
    id: int
        The template id to be run.
    arguments: dict
        Dictionary of arguments to be passed to the template.
    JSONValue: bool, optional
        If True, will return the JSON output of the template.
        If False, will return the file ids associated with the
        output results.
    client: :class:`civis.APIClient`, optional
        If not provided, an :class:`civis.APIClient` object will be
        created from the :envvar:`CIVIS_API_KEY`.

    Returns
    -------
    output: dict
        If JSONValue = False, dictionary of file ids with the keys
        being their output names.
        If JSONValue = True, JSON dict containing the results of the
        template run. Expects only a single JSON result. Will return
        nothing if either there is no JSON result or there is more
        than 1 JSON result.

    Examples
    --------
    >>> # Run template to return file_ids
    >>> run_template(my_template_id, arguments=my_dict_of_args)
    {'output': 1234567}
    >>> # Run template to return JSON output
    >>> run_template(my_template_id, arguments=my_dict_of_args, JSONValue=True)
    {'result1': 'aaa', 'result2': 123}
    """
    if client is None:
        client = APIClient()
    job = client.scripts.post_custom(id, arguments=arguments)
    run = client.scripts.post_custom_runs(job.id)
    fut = CivisFuture(
        client.scripts.get_custom_runs, (job.id, run.id), client=client
    )
    fut.result()
    outputs = client.scripts.list_custom_runs_outputs(job.id, run.id)
    if JSONValue:
        json_output = [
            o.value for o in outputs if o.object_type == "JSONValue"
        ]
        if len(json_output) == 0:
            log.warning("No JSON output for template {}".format(id))
            return
        if len(json_output) > 1:
            log.warning(
                "More than 1 JSON output for template {}"
                " -- returning only the first one.".format(id)
            )
        # Note that the cast to a dict is to convert
        # an expected Response object.
        return dict(json_output[0])
    else:
        file_ids = {o.name: o.object_id for o in outputs}
        return file_ids



from civis import _deprecation

import pytest


def adder(param1, param2=0, param3=0):
    return param1 + param2 + param3


def test_deprecate_kwarg():
    # Verify that we get a warning if the deprecated parameter is
    # used as a keyword argument.
    decorated_func = _deprecation.deprecate_param('v2.0.0', 'param2')(adder)

    with pytest.warns(FutureWarning) as record:
        output = decorated_func(1, param2=3, param3=5)

    assert output == 9, "The function should still give the expected output."
    assert len(record) == 1, "Only one warning should be raised."
    assert "v2.0.0" in record[0].message.args[0], \
        "The warning should mention the removal version."
    assert "param2" in record[0].message.args[0], \
        "The warning should mention the deprecated parameter."
    assert __name__ + ".adder" in record[0].message.args[0], \
        "The warning should mention the function name."


def test_deprecate_multiple_kwarg():
    # Verify that we get a warning if the deprecated parameter is
    # used as a keyword argument.
    decorated_func = _deprecation.deprecate_param(
        'v2.0.0', 'param2', 'param3')(adder)

    with pytest.warns(FutureWarning) as record:
        output = decorated_func(1, param2=3, param3=5)

    assert output == 9, "The function should still give the expected output."
    assert len(record) == 1, "Only one warning should be raised."
    assert "v2.0.0" in record[0].message.args[0], \
        "The warning should mention the removal version."
    assert "param2" in record[0].message.args[0], \
        "The warning should mention the first deprecated parameter."
    assert "param3" in record[0].message.args[0], \
        "The warning should mention the second deprecated parameter."
    assert __name__ + ".adder" in record[0].message.args[0], \
        "The warning should mention the function name."


def test_deprecate_pos_arg():
    # Verify that we get a warning if the deprecated parameter is
    # used as a positional argument.
    decorated_func = _deprecation.deprecate_param('v2.0.0', 'param2')(adder)

    with pytest.warns(FutureWarning) as record:
        output = decorated_func(1, 3, 5)

    assert output == 9, "The function should still give the expected output."
    assert len(record) == 1, "Only one warning should be raised."
    assert "v2.0.0" in record[0].message.args[0], \
        "The warning should mention the removal version."
    assert "param2" in record[0].message.args[0], \
        "The warning should mention the deprecated parameter."
    assert __name__ + ".adder" in record[0].message.args[0], \
        "The warning should mention the function name."


def test_deprecate_multiple_pos_arg():
    # Verify that we get a warning if the deprecated parameter is
    # used as a positional argument.
    decorated_func = _deprecation.deprecate_param(
        'v2.0.0', 'param2', 'param3')(adder)

    with pytest.warns(FutureWarning) as record:
        output = decorated_func(1, 3, 5)

    assert output == 9, "The function should still give the expected output."
    assert len(record) == 1, "Only one warning should be raised."
    assert "v2.0.0" in record[0].message.args[0], \
        "The warning should mention the removal version."
    assert "param2" in record[0].message.args[0], \
        "The warning should mention the first deprecated parameter."
    assert "param3" in record[0].message.args[0], \
        "The warning should mention the second deprecated parameter."
    assert __name__ + ".adder" in record[0].message.args[0], \
        "The warning should mention the function name."


def test_deprecate_no_warning():
    # Verify that we don't see a warning if we don't use the
    # deprecated parameter.
    decorated_func = _deprecation.deprecate_param('v2.0.0', 'param2')(adder)

    with pytest.warns(None) as record:
        output = decorated_func(1, param3=5)

    assert output == 6, "The function should still give the expected output."
    assert len(record) == 0, "No warnings should be raised."
